from beam.registry import default_registry, unregister
from beam.viewsets import ViewSet
from django.test import TestCase
from testapp.models import Dragonfly, Petaluridae
from testapp.views import DragonflyViewSet


class RegistryTest(TestCase):
    def test_registry_contains_viewset(self):
        self.assertIs(default_registry["testapp"]["dragonfly"], DragonflyViewSet)

    def test_duplicate_registration_warns(self):
        with self.assertWarns(UserWarning):

            class AnotherDragonFlyViewSet(DragonflyViewSet):
                pass

    def test_duplicate_registration_after_unregister_is_okay(self):
        custom_registry = {}

        class OneDragonFlyViewSet(DragonflyViewSet):
            registry = custom_registry

        unregister(custom_registry, Dragonfly)

        class AnotherDragonFlyViewSet(DragonflyViewSet):
            registry = custom_registry

        self.assertIs(custom_registry["testapp"]["dragonfly"], AnotherDragonFlyViewSet)

    def test_unregister_removes_app_and_model(self):
        custom_registry = {}

        class CustomDragonflyViewSet(ViewSet):
            registry = custom_registry
            model = Dragonfly

        class PetaluridaeViewSet(ViewSet):
            registry = custom_registry
            model = Petaluridae

        unregister(custom_registry, Dragonfly)

        assert "dragonfly" not in custom_registry["testapp"]

        unregister(custom_registry, Petaluridae)

        assert "testapp" not in custom_registry

    def test_register_with_custom_registry(self):
        custom_registry = {}

        class AnotherDragonFlyViewSet(DragonflyViewSet):
            registry = custom_registry

        self.assertIs(default_registry["testapp"]["dragonfly"], DragonflyViewSet)
        self.assertIs(custom_registry["testapp"]["dragonfly"], AnotherDragonFlyViewSet)



from unittest import TestCase, mock

from django.contrib.auth import get_user_model
from django.contrib.auth.models import Permission
from django.urls import reverse
from django_webtest import WebTest
from testapp.models import (
    CascadingSighting,
    Dragonfly,
    ProtectedSighting,
    Sighting,
    SightingReference,
)
from testapp.views import DragonflyViewSet, ExtraView, SightingViewSet

from beam.views import CreateView, DetailView, ListView, UpdateView


def user_with_perms(perms, username="foo", password="bar", user_model=None):
    user_model = user_model or get_user_model()
    user = user_model.objects.create_user(username=username, password=password)
    for perm in perms:
        app_label, codename = perm.split(".")
        try:
            permission = Permission.objects.get(
                content_type__app_label=app_label, codename=codename
            )
        except Permission.DoesNotExist as e:
            raise Permission.DoesNotExist(
                f"Could not find permission {app_label}.{codename}"
            ) from e
        user.user_permissions.add(permission)
    return user


class TemplateNameTest(TestCase):
    def test_detail_templates(self):
        view = DetailView(
            component=DragonflyViewSet().components["detail"], object=Dragonfly(pk=123)
        )
        self.assertEqual(
            view.get_template_names(),
            ["testapp/dragonfly_detail.html", "beam/detail.html"],
        )

    def test_list_templates(self):
        view = ListView(
            component=DragonflyViewSet().components["list"],
            object_list=Dragonfly.objects.none(),
        )
        self.assertEqual(
            view.get_template_names(), ["testapp/dragonfly_list.html", "beam/list.html"]
        )

    def test_update_templates(self):
        view = UpdateView(
            component=DragonflyViewSet().components["update"], object=Dragonfly(pk=123)
        )
        self.assertEqual(
            view.get_template_names(),
            [
                "testapp/dragonfly_update.html",
                "testapp/dragonfly_form.html",
                "beam/update.html",
            ],
        )

    def test_create_templates(self):
        view = CreateView(
            component=DragonflyViewSet().components["create"], object=Dragonfly(pk=123)
        )
        self.assertEqual(
            view.get_template_names(),
            [
                "testapp/dragonfly_create.html",
                "testapp/dragonfly_form.html",
                "beam/create.html",
            ],
        )

    def test_extra_templates(self):
        view = ExtraView(
            component=DragonflyViewSet().components["extra"], object=Dragonfly(pk=123)
        )
        self.assertEqual(
            view.get_template_names(),
            [
                "testapp/dragonfly_extra.html",
                "testapp/dragonfly_detail.html",
                "beam/detail.html",
            ],
        )

    def test_explicit_template_takes_precedence(self):
        with mock.patch.object(ExtraView, "template_name", "explicit.html"):
            view = ExtraView(
                component=DragonflyViewSet().components["extra"],
                object=Dragonfly(pk=123),
            )
            self.assertEqual(
                view.get_template_names(),
                [
                    "explicit.html",
                    "testapp/dragonfly_extra.html",
                    "beam/detail.html",
                ],
            )


class ViewTest(WebTest):
    def test_list(self):
        Dragonfly.objects.create(name="alpha", age=12)
        Dragonfly.objects.create(name="omega", age=99)
        response = self.app.get(
            DragonflyViewSet().links["list"].reverse(),
            user=user_with_perms(["testapp.view_dragonfly"]),
        )
        self.assertContains(response, "alpha")
        self.assertContains(response, "omega")

    def test_list_pagination_small_numbers(self):
        for i in range(DragonflyViewSet.list_paginate_by * 3):
            Dragonfly.objects.create(name="dragonfly-{:02}".format(i), age=100)

        first_page = self.app.get(
            DragonflyViewSet().links["list"].reverse(),
            user=user_with_perms(["testapp.view_dragonfly"]),
        )
        self.assertContains(first_page, "dragonfly-00")
        self.assertContains(first_page, "dragonfly-01")
        self.assertContains(first_page, "dragonfly-02")
        self.assertContains(first_page, "dragonfly-03")
        self.assertContains(first_page, "dragonfly-04")
        self.assertNotContains(first_page, "dragonfly-05")

        second_page = first_page.click("2")
        self.assertNotContains(second_page, "dragonfly-04")
        self.assertContains(second_page, "dragonfly-05")
        self.assertContains(second_page, "dragonfly-06")
        self.assertContains(second_page, "dragonfly-07")
        self.assertContains(second_page, "dragonfly-08")
        self.assertContains(second_page, "dragonfly-09")
        self.assertNotContains(second_page, "dragonfly-10")

    def test_list_pagination_large_numbers(self):
        for i in range(DragonflyViewSet.list_paginate_by * 10):
            Dragonfly.objects.create(name="dragonfly-{}".format(i), age=100)

        response = self.app.get(
            DragonflyViewSet().links["list"].reverse() + "?page=5",
            user=user_with_perms(["testapp.view_dragonfly"]),
        )

        # see beam/partials/pagination.html for a description
        self.assertContains(response, "page=1")
        self.assertContains(response, "page=2")
        self.assertNotContains(response, "page=3")

        self.assertContains(response, "page=4")
        self.assertContains(response, "page=6")

        self.assertNotContains(response, "page=8")
        self.assertContains(response, "page=9")
        self.assertContains(response, "page=10")

    def test_list_requires_permission(self):
        Dragonfly.objects.create(name="alpha", age=12)
        self.app.get(
            DragonflyViewSet().links["list"].reverse(),
            user=user_with_perms([]),
            status=403,
        )

    def test_list_redirects_on_login_required(self):
        Dragonfly.objects.create(name="alpha", age=12)
        response = self.app.get(
            DragonflyViewSet().links["list"].reverse(),
            user=None,
        )
        self.assertRedirects(
            response, "/accounts/login/?next=/dragonfly/", fetch_redirect_response=False
        )

    def test_list_search(self):
        Dragonfly.objects.create(name="alpha", age=12)
        Dragonfly.objects.create(name="omega", age=99)
        response = self.app.get(
            DragonflyViewSet().links["list"].reverse() + "?q=alpha",
            user=user_with_perms(["testapp.view_dragonfly"]),
        )
        self.assertContains(response, "alpha")
        self.assertNotContains(response, "omega")

    def test_list_filter_with_class(self):
        Dragonfly.objects.create(name="alpha", age=12)
        Dragonfly.objects.create(name="omega", age=99)
        response = self.app.get(
            DragonflyViewSet().links["list"].reverse(),
            user=user_with_perms(["testapp.view_dragonfly"]),
        )

        filter_form = response.forms["filter-form"]

        filter_form["filter-max_age"] = "0"
        max_0 = filter_form.submit()
        self.assertNotContains(max_0, "alpha")
        self.assertNotContains(max_0, "omega")

        filter_form["filter-max_age"] = "50"
        max_50 = filter_form.submit()
        self.assertContains(max_50, "alpha")
        self.assertNotContains(max_50, "omega")

        filter_form["filter-max_age"] = "100"
        max_100 = filter_form.submit()
        self.assertContains(max_100, "alpha")
        self.assertContains(max_100, "omega")

    def test_list_filter_with_fields(self):
        alpha = Dragonfly.objects.create(name="alpha", age=12)
        omega = Dragonfly.objects.create(name="omega", age=99)
        Sighting.objects.create(name="Berlin", dragonfly=alpha)
        Sighting.objects.create(name="Tokyo", dragonfly=omega)

        response = self.app.get(
            SightingViewSet().links["list"].reverse(),
            user=user_with_perms(["testapp.view_sighting"]),
        )

        filter_form = response.forms["filter-form"]

        filter_form["filter-name"] = "Tokyo"
        tokyo_response = filter_form.submit()
        self.assertNotContains(tokyo_response, "alpha")
        self.assertContains(tokyo_response, "omega")

        filter_form["filter-name"] = "nothing"
        empty_response = filter_form.submit()
        self.assertNotContains(empty_response, "alpha")
        self.assertNotContains(empty_response, "omega")
        self.assertContains(
            empty_response,
            "Could not find any sightings that match the current filters",
        )

    def test_list_sort(self):
        user = user_with_perms(["testapp.view_dragonfly"])

        Dragonfly.objects.create(name="alpha", age=12)
        Dragonfly.objects.create(name="omega", age=99)

        response = self.app.get(
            DragonflyViewSet().links["list"].reverse() + "?o=-name",
            user=user,
        )
        self.assertGreater(
            response.content.index(b"alpha"), response.content.index(b"omega")
        )

        response = self.app.get(
            DragonflyViewSet().links["list"].reverse() + "?o=name", user=user
        )
        self.assertLess(
            response.content.index(b"alpha"), response.content.index(b"omega")
        )

    def test_detail(self):
        alpha = Dragonfly.objects.create(name="alpha", age=47)
        Sighting.objects.create(name="Berlin", dragonfly=alpha)
        Sighting.objects.create(name="Paris", dragonfly=alpha)

        links = DragonflyViewSet().links

        response = self.app.get(
            links["detail"].reverse(alpha),
            user=(
                user_with_perms(
                    [
                        "testapp.view_dragonfly",
                        "testapp.change_dragonfly",
                        "testapp.delete_dragonfly",
                    ],
                )
            ),
        )

        self.assertContains(response, "alpha")
        self.assertContains(response, "Title of sightings")

        self.assertContains(response, "Berlin")
        self.assertContains(response, "Paris")

        self.assertContains(response, 'href="{}"'.format(links["list"].reverse(alpha)))
        self.assertContains(
            response, 'href="{}"'.format(links["update"].reverse(alpha))
        )
        self.assertContains(
            response, 'href="{}"'.format(links["delete"].reverse(alpha))
        )

    def test_detail_requires_permission(self):
        alpha = Dragonfly.objects.create(name="alpha", age=47)
        links = DragonflyViewSet().links
        self.app.get(
            links["detail"].reverse(alpha),
            user=user_with_perms([]),
            status=403,
        )

    def test_detail_links_related(self):
        alpha = Dragonfly.objects.create(name="alpha", age=47)
        sighting = Sighting.objects.create(dragonfly=alpha)
        links = SightingViewSet().links
        detail_page = self.app.get(
            links["detail"].reverse(sighting),
            user=user_with_perms(
                [
                    "testapp.view_dragonfly",
                    "testapp.view_sighting",
                ]
            ),
        )
        detail_page.click(
            "alpha", href=DragonflyViewSet().links["detail"].reverse(alpha)
        )

    def test_detail_links_to_related_require_permission(self):
        alpha = Dragonfly.objects.create(name="alpha", age=47)
        sighting = Sighting.objects.create(dragonfly=alpha)
        links = SightingViewSet().links
        detail_page = self.app.get(
            links["detail"].reverse(sighting),
            user=user_with_perms(
                [
                    "testapp.view_sighting",
                ]
            ),
        )
        with self.assertRaises(IndexError):
            detail_page.click(
                "alpha", href=DragonflyViewSet().links["detail"].reverse(alpha)
            )

    def test_update(self):
        alpha = Dragonfly.objects.create(name="alpha", age=47)
        response = self.app.get(
            DragonflyViewSet().links["update"].reverse(alpha),
            user=user_with_perms(
                [
                    "testapp.view_dragonfly",
                    "testapp.change_dragonfly",
                ]
            ),
        )
        self.assertContains(response, "alpha")
        self.assertIn("form", response.context)
        self.assertEqual(response.context["form"]["name"].value(), "alpha")

        form = response.form
        form["name"] = "first"
        update_response = form.submit()

        self.assertRedirects(
            update_response,
            DragonflyViewSet().links["detail"].reverse(obj=alpha),
            fetch_redirect_response=False,
        )

        self.assertContains(
            update_response.follow(),
            "The dragonfly &quot;first&quot; was changed successfully.",
        )

    def test_update_requires_permission(self):
        alpha = Dragonfly.objects.create(name="alpha", age=47)
        self.app.get(
            DragonflyViewSet().links["update"].reverse(alpha),
            user=user_with_perms([]),
            status=403,
        )

    def test_extra_component_requires_permission(self):
        alpha = Dragonfly.objects.create(name="alpha", age=47)
        self.app.get(
            (
                DragonflyViewSet()
                .links["extra"]
                .reverse(alpha, override_kwargs={"special": "param"})
            ),
            user=user_with_perms([]),
            status=403,
        )

    def test_update_and_continue_editing(self):
        alpha = Dragonfly.objects.create(name="alpha", age=47)
        links = DragonflyViewSet().links

        response = self.app.get(
            links["update"].reverse(alpha),
            user=user_with_perms(
                [
                    "testapp.view_dragonfly",
                    "testapp.change_dragonfly",
                ]
            ),
        )
        form = response.form
        update_response = form.submit("submit", value="save_and_continue_editing")

        self.assertRedirects(
            update_response,
            DragonflyViewSet().links["update"].reverse(obj=alpha),
            fetch_redirect_response=False,
        )

    def test_delete(self):
        alpha = Dragonfly.objects.create(name="alpha", age=47)
        delete_url = DragonflyViewSet().links["delete"].reverse(alpha)
        response = self.app.get(
            delete_url,
            user=user_with_perms(
                ["testapp.view_dragonfly", "testapp.delete_dragonfly"]
            ),
        )

        self.assertContains(response, "Are you sure you want to delete")
        self.assertContains(response, "alpha")

        delete_response = response.form.submit()

        self.assertRedirects(
            delete_response,
            DragonflyViewSet().links["list"].reverse(),
            fetch_redirect_response=False,
        )
        self.assertContains(
            delete_response.follow(),
            "The dragonfly &quot;alpha&quot; was deleted successfully",
        )
        self.assertFalse(Dragonfly.objects.filter(name="alpha").exists())

    def test_delete_shows_related(self):
        alpha = Dragonfly.objects.create(name="alpha", age=47)
        CascadingSighting.objects.create(dragonfly=alpha, name="A related sighting")

        delete_url = DragonflyViewSet().links["delete"].reverse(alpha)
        response = self.app.get(
            delete_url,
            user=user_with_perms(["testapp.delete_dragonfly"]),
        )
        self.assertContains(response, "The following objects")
        self.assertContains(response, "sighting")

        delete_response = response.form.submit()

        self.assertRedirects(
            delete_response,
            DragonflyViewSet().links["list"].reverse(),
            fetch_redirect_response=False,
        )

        self.assertFalse(Dragonfly.objects.filter(name="alpha").exists())

    def test_delete_protected_not_allowed(self):
        user = user_with_perms(["testapp.delete_dragonfly"])
        alpha = Dragonfly.objects.create(name="alpha", age=47)
        ProtectedSighting.objects.create(dragonfly=alpha, name="A related sighting")

        delete_url = DragonflyViewSet().links["delete"].reverse(alpha)
        response = self.app.get(
            delete_url,
            user=user,
        )
        self.assertContains(response, "You can't delete")
        self.assertContains(response, "the following objects depend")
        self.assertContains(response, "sighting")

        self.app.post(delete_url, user=user, status=403)

        assert Dragonfly.objects.filter(name="alpha").exists()

    def test_delete_requires_permission(self):
        user = user_with_perms([])
        alpha = Dragonfly.objects.create(name="alpha", age=47)
        delete_url = DragonflyViewSet().links["delete"].reverse(alpha)

        self.app.get(delete_url, user=user, status=403)
        self.app.post(delete_url, user=user, status=403)

    def test_create_with_inlines(self):
        user = user_with_perms(["testapp.view_dragonfly", "testapp.add_dragonfly"])
        create_page = self.app.get(
            DragonflyViewSet().links["create"].reverse(), user=user
        )

        form = create_page.form

        form["name"] = "foobar"
        form["age"] = "81"
        form["sighting_set-0-name"] = "Tokyo"

        response = form.submit()

        dragonfly = Dragonfly.objects.get(name="foobar")

        self.assertRedirects(
            response,
            DragonflyViewSet().links["detail"].reverse(dragonfly),
            fetch_redirect_response=False,
        )

        self.assertContains(
            response.follow(),
            "The dragonfly &quot;foobar&quot; was added successfully.",
        )
        self.assertEqual(dragonfly.age, 81)
        self.assertEqual(dragonfly.sighting_set.get().name, "Tokyo")

    def test_only_popup_param_is_preserved_in_detail_links(self):
        user = user_with_perms(["testapp.view_dragonfly", "testapp.change_dragonfly"])

        instance = Dragonfly.objects.create(name="alpha", age=12)
        response = self.app.get(
            DragonflyViewSet().links["list"].reverse(),
            user=user,
            params={"_popup": "id_test", "not_preserved": "nope"},
        )

        detail_url = DragonflyViewSet().links["update"].reverse(instance)

        self.assertContains(response, detail_url)
        self.assertContains(response, detail_url + "?_popup=id_test")
        self.assertNotContains(response, detail_url + "?_popup=id_test&not_preserved")

    def test_custom_detail_template_for_inline(self):
        user = user_with_perms(["testapp.view_dragonfly"])
        dragonfly = Dragonfly.objects.create(name="alpha", age=12)
        response = self.app.get(
            DragonflyViewSet().links["detail"].reverse(dragonfly),
            user=user,
        )
        self.assertContains(response, "Awesome extra custom template title")

    def test_custom_form_template_for_inline(self):
        user = user_with_perms(["testapp.view_dragonfly", "testapp.add_dragonfly"])
        response = self.app.get(
            DragonflyViewSet().links["create"].reverse(),
            user=user,
        )
        self.assertContains(response, "Awesome extra custom template title")

    def test_detail_filter_inlines(self):
        user = user_with_perms(["testapp.view_dragonfly"])
        dragonfly = Dragonfly.objects.create(name="alpha", age=12)

        Sighting.objects.create(name="sighting-one", dragonfly=dragonfly)
        Sighting.objects.create(name="sighting-two", dragonfly=dragonfly)

        response = self.app.get(
            DragonflyViewSet().links["detail"].reverse(dragonfly),
            user=user,
        )

        self.assertContains(response, "sighting-one")
        self.assertContains(response, "sighting-two")

        response.form["sighting_set-filter-name"] = "sighting-one"
        filtered = response.form.submit()

        self.assertContains(filtered, "sighting-one")
        self.assertNotContains(filtered, "sighting-two")

    def test_detail_filter_multiple_inlines(self):
        user = user_with_perms(["testapp.view_dragonfly"])
        dragonfly = Dragonfly.objects.create(name="alpha", age=12)

        Sighting.objects.create(name="regular-sighting-one", dragonfly=dragonfly)
        Sighting.objects.create(name="regular-sighting-two", dragonfly=dragonfly)

        ProtectedSighting.objects.create(
            name="protected-sighting-one", dragonfly=dragonfly
        )
        ProtectedSighting.objects.create(
            name="protected-sighting-two", dragonfly=dragonfly
        )

        response = self.app.get(
            DragonflyViewSet().links["detail"].reverse(dragonfly),
            user=user,
        )

        self.assertContains(response, "regular-sighting-one")
        self.assertContains(response, "regular-sighting-two")

        self.assertContains(response, "protected-sighting-one")
        self.assertContains(response, "protected-sighting-two")

        form = response.forms["sighting_set-filter-form"]
        form["sighting_set-filter-name"] = "regular-sighting-one"
        filtered = form.submit()

        self.assertContains(filtered, "regular-sighting-one")
        self.assertNotContains(filtered, "regular-sighting-two")

        self.assertContains(filtered, "protected-sighting-one")
        self.assertContains(filtered, "protected-sighting-two")

    def test_tabular_inline_detail(self):
        user = user_with_perms(["testapp.view_dragonfly"])
        dragonfly = Dragonfly.objects.create(name="alpha", age=12)

        CascadingSighting.objects.create(name="sighting-one", dragonfly=dragonfly)
        CascadingSighting.objects.create(name="sighting-two", dragonfly=dragonfly)

        response = self.app.get(
            DragonflyViewSet().links["detail"].reverse(dragonfly),
            user=user,
        )

        self.assertContains(response, "sighting-one")
        self.assertContains(response, "sighting-two")
        self.assertContains(response, "cascadingsighting_set-table")

    def test_tabular_inline_form(self):
        user = user_with_perms(["testapp.view_dragonfly", "testapp.change_dragonfly"])
        dragonfly = Dragonfly.objects.create(name="alpha", age=12)

        CascadingSighting.objects.create(name="sighting-one", dragonfly=dragonfly)
        CascadingSighting.objects.create(name="sighting-two", dragonfly=dragonfly)

        response = self.app.get(
            DragonflyViewSet().links["update"].reverse(dragonfly),
            user=user,
        )

        self.assertContains(response, "sighting-one")
        self.assertContains(response, "sighting-two")
        self.assertContains(response, "cascadingsighting_set-table")

    def test_delete_inline(self):
        user = user_with_perms(["testapp.view_dragonfly", "testapp.change_dragonfly"])

        dragonfly = Dragonfly.objects.create(name="alpha", age=12)
        Sighting.objects.create(name="a-sighting", dragonfly=dragonfly)

        response = self.app.get(
            DragonflyViewSet().links["update"].reverse(dragonfly),
            user=user,
        )

        self.assertContains(response, "a-sighting")

        form = response.form
        form["sighting_set-0-DELETE"] = True
        delete_response = form.submit()

        self.assertRedirects(
            delete_response, DragonflyViewSet().links["detail"].reverse(dragonfly)
        )

        self.assertFalse(dragonfly.sighting_set.exists())

    def test_delete_protected_inline(self):
        user = user_with_perms(["testapp.view_dragonfly", "testapp.change_dragonfly"])

        dragonfly = Dragonfly.objects.create(name="alpha", age=12)
        sighting = Sighting.objects.create(name="a-sighting", dragonfly=dragonfly)
        SightingReference.objects.create(sighting=sighting)

        response = self.app.get(
            DragonflyViewSet().links["update"].reverse(dragonfly),
            user=user,
        )

        self.assertContains(response, "a-sighting")

        form = response.form
        form["sighting_set-0-DELETE"] = True
        delete_response = form.submit()

        self.assertContains(
            delete_response,
            "would require deleting the following protected related objects",
        )

        self.assertTrue(dragonfly.sighting_set.exists())

    def test_navigation(self):
        user = user_with_perms(["testapp.view_dragonfly", "testapp.view_sighting"])
        base_page = self.app.get(reverse("base-template"), user=user)
        self.assertContains(base_page, "Testapp")
        self.assertContains(base_page, DragonflyViewSet().links["list"].reverse())
        self.assertContains(base_page, SightingViewSet().links["list"].reverse())

    def test_navigation_respects_permissions(self):
        user = user_with_perms(["testapp.view_sighting"])
        base_page = self.app.get(reverse("base-template"), user=user)
        self.assertNotContains(base_page, DragonflyViewSet().links["list"].reverse())
        self.assertContains(base_page, SightingViewSet().links["list"].reverse())

    def test_navigation_hides_empty_submenus(self):
        user = user_with_perms([])
        base_page = self.app.get(reverse("base-template"), user=user)
        self.maxDiff = None
        self.assertNotContains(base_page, "Testapp")
        self.assertNotContains(base_page, SightingViewSet().links["list"].reverse())
        self.assertNotContains(base_page, DragonflyViewSet().links["list"].reverse())


class InlinePaginationTest(WebTest):
    def test_paginated_detail_inlines(self):
        user = user_with_perms(["testapp.view_dragonfly"])

        dragonfly = Dragonfly.objects.create(name="alpha", age=12)
        for i in range(6):  # SightingInline.paginate_by + 1
            Sighting.objects.create(
                name="sighting-at-{}".format(i), dragonfly=dragonfly
            )

        response = self.app.get(
            DragonflyViewSet().links["detail"].reverse(dragonfly),
            user=user,
        )

        self.assertContains(response, "sighting-at-0")
        self.assertContains(response, "sighting-at-1")
        self.assertContains(response, "sighting-at-2")
        self.assertContains(response, "sighting-at-3")
        self.assertContains(response, "sighting-at-4")
        self.assertNotContains(response, "sighting-at-5")

        next_page = response.click("Next")

        self.assertNotContains(next_page, "sighting-at-0")
        self.assertNotContains(next_page, "sighting-at-1")
        self.assertNotContains(next_page, "sighting-at-2")
        self.assertNotContains(next_page, "sighting-at-3")
        self.assertNotContains(next_page, "sighting-at-4")
        self.assertContains(next_page, "sighting-at-5")

    def test_paginated_update_inlines(self):
        user = user_with_perms(["testapp.view_dragonfly", "testapp.change_dragonfly"])

        dragonfly = Dragonfly.objects.create(name="alpha", age=12)
        for i in range(6):  # SightingInline.paginate_by + 1
            Sighting.objects.create(
                name="sighting-at-{}".format(i), dragonfly=dragonfly
            )

        response = self.app.get(
            DragonflyViewSet().links["update"].reverse(dragonfly),
            user=user,
        )

        self.assertContains(response, "sighting-at-0")
        self.assertContains(response, "sighting-at-1")
        self.assertContains(response, "sighting-at-2")
        self.assertContains(response, "sighting-at-3")
        self.assertContains(response, "sighting-at-4")
        self.assertNotContains(response, "sighting-at-5")

        form = response.form
        form["sighting_set-0-name"] = "sighting-at-0-changed"
        form["sighting_set-1-DELETE"] = True
        update_response = form.submit()
        self.assertRedirects(
            update_response, DragonflyViewSet().links["detail"].reverse(dragonfly)
        )

        self.assertSetEqual(
            set(dragonfly.sighting_set.values_list("name", flat=True)),
            {
                "sighting-at-0-changed",
                "sighting-at-2",
                "sighting-at-3",
                "sighting-at-4",
                "sighting-at-5",
            },
        )

    def test_dashboard_no_permissions(self):
        user_no_perms = user_with_perms([], username="user-no-perms")

        dashboard_no_perms = self.app.get(
            reverse("dashboard"),
            user=user_no_perms,
        )

        self.assertNotContains(
            dashboard_no_perms, Dragonfly._meta.verbose_name_plural.capitalize()
        )
        self.assertNotContains(dashboard_no_perms, 'href="/dragonfly/"')
        self.assertNotContains(dashboard_no_perms, 'href="/dragonfly/create/"')

        self.assertNotContains(
            dashboard_no_perms, Sighting._meta.verbose_name_plural.capitalize()
        )
        self.assertNotContains(dashboard_no_perms, 'href="/sighting/"')
        self.assertNotContains(dashboard_no_perms, 'href="/sighting/create/"')

    def test_dashboard_one_app_only(self):
        user_dragonfly_only = user_with_perms(
            [
                "testapp.view_dragonfly",
                "testapp.add_dragonfly",
            ],
            username="user-dragonfly",
        )

        dashboard_dragonfly_only = self.app.get(
            reverse("dashboard"),
            user=user_dragonfly_only,
        )

        self.assertContains(
            dashboard_dragonfly_only, Dragonfly._meta.verbose_name_plural.capitalize()
        )
        self.assertContains(dashboard_dragonfly_only, 'href="/dragonfly/"')
        self.assertContains(dashboard_dragonfly_only, 'href="/dragonfly/create/"')

        self.assertNotContains(
            dashboard_dragonfly_only, Sighting._meta.verbose_name_plural.capitalize()
        )
        self.assertNotContains(dashboard_dragonfly_only, 'href="/sighting/"')
        self.assertNotContains(dashboard_dragonfly_only, 'href="/sighting/create/"')

    def test_dashboard_view_all_permissions(self):
        user_view_all = user_with_perms(
            [
                "testapp.view_dragonfly",
                "testapp.view_sighting",
            ],
            username="user-view-all",
        )

        dashboard_view_all = self.app.get(
            reverse("dashboard"),
            user=user_view_all,
        )
        self.assertContains(
            dashboard_view_all, Dragonfly._meta.verbose_name_plural.capitalize()
        )
        self.assertContains(dashboard_view_all, 'href="/dragonfly/"')
        self.assertNotContains(dashboard_view_all, 'href="/dragonfly/create/"')

        self.assertContains(
            dashboard_view_all, Sighting._meta.verbose_name_plural.capitalize()
        )
        self.assertContains(dashboard_view_all, 'href="/sighting/"')
        self.assertNotContains(dashboard_view_all, 'href="/sighting/create/"')



from beam import RelatedInline
from django.test import TestCase
from testapp.models import Dragonfly, Sighting


class InlineTest(TestCase):
    def test_inline_formset_is_generated(self):
        class SightingInline(RelatedInline):
            title = "Title of sightings"
            fields = ["name"]
            model = Sighting
            foreign_key_field = "dragonfly"

        formset = SightingInline(parent_model=Dragonfly).formset

        self.assertIn("name", formset.forms[0].fields)

    def test_inlines_prefix_works_without_formset(self):
        class SightingInline(RelatedInline):
            title = "Title of sightings"
            fields = ["name", "created_at"]
            model = Sighting
            foreign_key_field = "dragonfly"

        self.assertEqual(SightingInline(parent_model=Dragonfly).prefix, "sighting_set")



from django.contrib.messages.middleware import MessageMiddleware
from django.contrib.messages.storage.fallback import FallbackStorage
from django.contrib.sessions.middleware import SessionMiddleware
from django.core.exceptions import PermissionDenied
from django.http.response import HttpResponse
from django.test import RequestFactory, TestCase
from django.test.utils import override_settings
from django.urls import include, path
from reversion import is_registered, set_comment
from reversion.models import Version
from test_views import user_with_perms
from testapp.models import Dragonfly, Petaluridae, Sighting

from beam import RelatedInline
from beam.contrib.reversion.viewsets import VersionViewSet
from beam.registry import RegistryType

registry: RegistryType = {}


class GetResponse:
    def __call__(self):
        return HttpResponse()


class SightingInline(RelatedInline):
    title = "Title of sightings"
    fields = ["name"]
    model = Sighting
    foreign_key_field = "dragonfly"


class VersionedDragonflyViewSet(VersionViewSet):
    registry = registry

    inline_classes = [SightingInline]
    model = Dragonfly
    list_search_fields = ["name"]
    fields = ["name", "age"]


urlpatterns = [
    path("dragonfly/", include(VersionedDragonflyViewSet().get_urls())),
]

VersionedDragonflyViewSet()


class RegistrationTest(TestCase):
    def test_versioned_view_set_models_are_registered(self):
        self.assertTrue(is_registered(Dragonfly))

    def test_inlines_of_versioned_view_set_models_are_registered(self):
        self.assertTrue(is_registered(Sighting))

    def test_unrelated_models_are_not_registered(self):
        self.assertFalse(is_registered(Petaluridae))


class ReversionViewTest(TestCase):
    # FIXME: this should probably be a regular webtest with the urlconf above
    def test_using_create_view_creates_a_revision(self):
        user = user_with_perms(["testapp.add_dragonfly"])
        create_view = VersionedDragonflyViewSet()._get_view(
            VersionedDragonflyViewSet().components["create"]
        )

        request = RequestFactory().post(
            "/dragonfly/create/",
            data={
                "name": "versionfly",
                "age": 81,
                "sighting_set-TOTAL_FORMS": 1,
                "sighting_set-INITIAL_FORMS": 0,
                "sighting_set-MIN_NUM_FORMS": 0,
                "sighting_set-MAX_NUM_FORMS": 10,
                "sighting_set-0-name": "Tokyo",
            },
        )
        request.user = user

        middleware = SessionMiddleware(get_response=GetResponse())
        middleware.process_request(request)
        messages = FallbackStorage(request)
        setattr(request, "_messages", messages)

        response = create_view(request)

        self.assertEqual(response.status_code, 302)

        fly = Dragonfly.objects.get()
        fly_version = Version.objects.get_for_object_reference(Dragonfly, fly.pk).get()

        self.assertEqual(fly_version.object.name, "versionfly")
        self.assertEqual(fly_version.field_dict["name"], "versionfly")
        self.assertEqual(fly_version.revision.user, user)
        self.assertEqual(fly_version.revision.comment, "create")

        sighting_version = Version.objects.get_for_object_reference(
            Sighting, fly.sighting_set.get().pk
        ).get()

        self.assertEqual(sighting_version.field_dict["name"], "Tokyo")

    def test_using_update_view_creates_a_revision(self):
        user = user_with_perms(["testapp.change_dragonfly"])
        alpha = Dragonfly.objects.create(name="alpha", age=47)

        update_view = VersionedDragonflyViewSet()._get_view(
            VersionedDragonflyViewSet().components["update"]
        )

        request = RequestFactory().post(
            "/dragonfly/update/{}/".format(alpha.pk),
            data={
                "name": "beta",
                "age": 81,
                "sighting_set-TOTAL_FORMS": 0,
                "sighting_set-INITIAL_FORMS": 0,
                "sighting_set-MIN_NUM_FORMS": 0,
                "sighting_set-MAX_NUM_FORMS": 10,
            },
        )
        request.user = user
        middleware = SessionMiddleware(get_response=GetResponse())
        middleware.process_request(request)
        messages = FallbackStorage(request)
        setattr(request, "_messages", messages)

        response = update_view(request, pk=alpha.pk)

        self.assertEqual(response.status_code, 302)

        fly_version = Version.objects.get_for_object_reference(
            Dragonfly, alpha.pk
        ).latest("revision__date_created")

        self.assertEqual(fly_version.object.name, "beta")
        self.assertEqual(fly_version.field_dict["name"], "beta")
        self.assertEqual(fly_version.revision.user, user)
        self.assertEqual(fly_version.revision.comment, "update")

    @override_settings(ROOT_URLCONF=__name__)
    def test_revision_is_visible_in_list(self):
        alpha = Dragonfly.objects.create(name="alpha", age=47)

        request = RequestFactory().get("/", {})
        request.user = user_with_perms(["testapp.view_dragonfly"])

        with VersionedDragonflyViewSet().create_revision(request):
            set_comment("number one")
            alpha.save()

        with VersionedDragonflyViewSet().create_revision(request):
            set_comment("number two")
            alpha.save()

        with VersionedDragonflyViewSet().create_revision(request):
            set_comment("number three")
            alpha.save()

        self.assertEqual(
            Version.objects.get_for_object_reference(Dragonfly, alpha.pk).count(), 3
        )

        view = VersionedDragonflyViewSet()._get_view(
            VersionedDragonflyViewSet().components["version_list"]
        )

        response = view(request, pk=alpha.pk)

        self.assertContains(response, "number one")
        self.assertContains(response, "number two")
        self.assertContains(response, "number three")

    def test_version_list_requires_view_permission(self):
        alpha = Dragonfly.objects.create(name="alpha", age=47)

        request = RequestFactory().get("/", {})
        request.user = user_with_perms([])

        with VersionedDragonflyViewSet().create_revision(request):
            set_comment("number one")
            alpha.save()

        view = VersionedDragonflyViewSet()._get_view(
            VersionedDragonflyViewSet().components["version_list"]
        )

        with self.assertRaises(PermissionDenied):
            view(request, pk=alpha.pk)

    @override_settings(ROOT_URLCONF=__name__)
    def test_show_detail_from_previous_version(self):
        alpha = Dragonfly.objects.create(name="alpha", age=47)
        sighting = Sighting.objects.create(name="Berlin", dragonfly=alpha)

        request = RequestFactory().get("/", {})
        request.user = user_with_perms(["testapp.view_dragonfly"])

        with VersionedDragonflyViewSet().create_revision(request):
            alpha.save()

        version = Version.objects.get_for_object_reference(Dragonfly, alpha.pk).latest(
            "revision__date_created"
        )

        alpha.name = "beta"
        alpha.save()

        sighting.name = "Tokyo"
        sighting.save()

        detail_view = VersionedDragonflyViewSet()._get_view(
            VersionedDragonflyViewSet().components["detail"]
        )
        detail_page = detail_view(request, pk=alpha.pk)

        self.assertContains(detail_page, "beta")
        self.assertContains(detail_page, "Tokyo")

        version_view = VersionedDragonflyViewSet()._get_view(
            VersionedDragonflyViewSet().components["version_detail"]
        )
        version_page = version_view(request, pk=alpha.pk, version_id=version.pk)

        self.assertContains(version_page, "alpha")
        self.assertContains(version_page, "Berlin")

        self.assertNotContains(version_page, "beta")
        self.assertNotContains(version_page, "Tokyo")

    def test_version_detail_requires_view_perm(self):
        alpha = Dragonfly.objects.create(name="alpha", age=47)
        Sighting.objects.create(name="Berlin", dragonfly=alpha)

        request = RequestFactory().get("/", {})
        request.user = user_with_perms([])

        with VersionedDragonflyViewSet().create_revision(request):
            alpha.save()

        Version.objects.get_for_object_reference(Dragonfly, alpha.pk).latest(
            "revision__date_created"
        )
        detail_view = VersionedDragonflyViewSet()._get_view(
            VersionedDragonflyViewSet().components["detail"]
        )
        with self.assertRaises(PermissionDenied):
            detail_view(request, pk=alpha.pk)

    def test_revert_to_previous_version(self):
        alpha = Dragonfly.objects.create(name="alpha", age=47)
        sighting = Sighting.objects.create(name="Berlin", dragonfly=alpha)

        request = RequestFactory().post("/", {})
        request.user = user_with_perms(["testapp.change_dragonfly"])
        SessionMiddleware(get_response=GetResponse()).process_request(request)
        MessageMiddleware(get_response=GetResponse()).process_request(request)

        with VersionedDragonflyViewSet().create_revision(request):
            alpha.save()

        version = Version.objects.get_for_object_reference(Dragonfly, alpha.pk).latest(
            "revision__date_created"
        )

        alpha.name = "beta"
        alpha.save()

        sighting.name = "Tokyo"
        sighting.save()

        version_view = VersionedDragonflyViewSet()._get_view(
            VersionedDragonflyViewSet().components["version_restore"]
        )
        response = version_view(request, pk=alpha.pk, version_id=version.pk)

        self.assertEqual(response.status_code, 302)

        alpha.refresh_from_db()
        sighting.refresh_from_db()

        self.assertEqual(alpha.name, "alpha")
        self.assertEqual(sighting.name, "Berlin")

    def test_revert_to_previous_version_requires_change_perm(self):
        alpha = Dragonfly.objects.create(name="alpha", age=47)
        Sighting.objects.create(name="Berlin", dragonfly=alpha)

        request = RequestFactory().post("/", {})
        request.user = user_with_perms([])

        with VersionedDragonflyViewSet().create_revision(request):
            alpha.save()

        version = Version.objects.get_for_object_reference(Dragonfly, alpha.pk).latest(
            "revision__date_created"
        )

        version_view = VersionedDragonflyViewSet()._get_view(
            VersionedDragonflyViewSet().components["version_restore"]
        )
        with self.assertRaises(PermissionDenied):
            version_view(request, pk=alpha.pk, version_id=version.pk)



import inspect
from typing import List, Mapping, Optional, Set, Type

import django_filters
from django.urls import reverse

from .actions import Action
from .utils import check_permission


class BaseComponent:
    show_link = True

    def __init__(
        self,
        viewset=None,
        name=None,
        verbose_name=None,
        url_name=None,
        url_kwargs=None,
        url_namespace=None,
        permission=None,
        resolve_url=None,
        **kwargs
    ):
        if not name:
            raise ValueError("Components need a name")

        self.viewset = viewset
        self.name = name
        self.verbose_name = verbose_name or self.name
        self.url_name = url_name
        self.url_kwargs = url_kwargs or {}
        self.resolve_url = resolve_url or reverse
        self.url_namespace = url_namespace

        if isinstance(permission, str):
            context = {"component": self}
            if getattr(self, "model", None) is not None:
                opts = self.model._meta
                context["model_name"] = opts.model_name
                context["app_label"] = opts.app_label
            self.permission = permission.format(**context)
        else:
            self.permission = permission

    def __repr__(self):
        if self.viewset:
            return "<{} {}.{} {}>".format(
                self.__class__.__name__,
                self.viewset.__module__,
                self.viewset.__class__.__name__,
                repr(self.name),
            )
        return "<{} {}>".format(self.__class__.__name__, repr(self.name))

    @classmethod
    def get_arguments(cls) -> Set[str]:
        """
        Get a list of arguments that can be passed from the viewset.

        These arguments will be used by ViewSet._get_components
        when instantiating components.
        """
        arguments = set()
        for class_ in inspect.getmro(cls):
            for arg in inspect.getfullargspec(class_).args[1:]:
                arguments.add(arg)
        return arguments

    def has_perm(self, user, obj=None, request=None, override_kwargs=None) -> bool:
        """
        Check whether a given user has access to this component.
        The optional parameters mirror those for reverse so that
        every possible url can be checked as well.

        :param obj: An optional model instance
        :param request: An optional request
        :param override_kwargs: An optional dict overriding url kwargs

        :return: A boolean, is the user allowed to interact with this component?
        """
        return check_permission(permission=self.permission, user=user, obj=obj)

    def reverse(self, obj=None, request=None, override_kwargs=None):
        """
        Get a url for this component.

        :param obj: An optional model instance
        :param request: An optional request
        :param override_kwargs: An optional dict overriding url kwargs
        :return: A url
        """
        kwargs = self.resolve_url_kwargs(obj, request, override_kwargs)
        url_name = self.url_name
        if self.url_namespace:
            url_name = self.url_namespace + ":" + url_name

        return self.resolve_url(url_name, kwargs=kwargs)

    def resolve_url_kwargs(self, obj=None, request=None, override_kwargs=None):
        """
        Used to build the url kwargs that Component.reverse passed to django's reverse.

        Keys with None values are dropped from the result so you can't have urls
        with None in them.

        :param obj: An optional model instance
        :param request: An optional request
        :param override_kwargs: An optional dict overriding url kwargs
        :return: A dict of url kwargs that can be passed to django's reverse.
        """
        override_kwargs = override_kwargs or {}
        kwargs = {}
        assert self.url_kwargs is not None, str(self.__dict__)
        for kwarg, name_or_callable in self.url_kwargs.items():
            if callable(name_or_callable):
                kwargs[kwarg] = name_or_callable(obj, request)
            elif hasattr(obj, name_or_callable):
                kwargs[kwarg] = getattr(obj, name_or_callable)
        kwargs.update(override_kwargs)

        return {k: v for k, v in kwargs.items() if v is not None}


class Component(BaseComponent):
    def __init__(
        self,
        view_class=None,
        fields=None,
        url=None,
        layout=None,
        queryset=None,
        model=None,
        inline_classes=None,
        link_layout=None,
        name=None,
        url_name=None,
        url_namespace=None,
        **kwargs
    ):
        self.url = url

        if not view_class:
            raise ValueError(
                "A component requires a view class, if you do "
                "not want to serve a view use BaseComponent"
            )
        self.view_class = view_class

        self.inline_classes = inline_classes

        self.fields = fields  # FIXME do the layout / fields auto-create thing
        self.layout = layout

        if model is None and queryset is None:
            raise ValueError(
                "Component {} needs at least one of model, queryset".format(name)
            )
        elif model is not None and queryset is None:
            queryset = model._default_manager
        elif queryset is not None and model is None:
            model = queryset.model

        self.model = model
        self.queryset = queryset

        if not url_name:
            url_name = "{}_{}_{}".format(
                self.model._meta.app_label, self.model._meta.model_name, name
            )

        self.link_layout = (
            link_layout if link_layout is not None else ["!{}".format(name), "..."]
        )
        super().__init__(
            name=name, url_name=url_name, url_namespace=url_namespace, **kwargs
        )

    @property
    def queryset(self):
        # ensure we don't keep stale copies of querysets
        return self._queryset.all()

    @queryset.setter
    def queryset(self, queryset):
        self._queryset = queryset


class FormComponent(Component):
    def __init__(self, form_class=None, **kwargs):
        self.form_class = form_class
        super().__init__(**kwargs)


class ListComponent(Component):
    def __init__(
        self,
        list_search_fields: Optional[List[str]] = None,
        list_paginate_by: Optional[int] = None,
        list_item_link_layout: Optional[List[str]] = None,
        list_sort_fields: Optional[List[str]] = None,
        list_sort_fields_columns: Optional[Mapping[str, str]] = None,
        list_filterset_fields: Optional[List[str]] = None,
        list_filterset_class: Optional[
            Type[django_filters.filterset.BaseFilterSet]
        ] = None,
        list_action_classes: Optional[List[Type[Action]]] = None,
        **kwargs
    ):
        self.list_search_fields = list_search_fields
        self.list_paginate_by = list_paginate_by
        self.list_item_link_layout = list_item_link_layout
        self.list_sort_fields = list_sort_fields
        self.list_sort_fields_columns = list_sort_fields_columns
        self.list_filterset_fields = list_filterset_fields
        self.list_filterset_class = list_filterset_class
        self.list_actions_classes = list_action_classes
        super().__init__(**kwargs)


class Link(BaseComponent):
    """
    A component class that can be added to ViewSet.links to add links to external views.
    """

    pass



from unittest import mock

from django.test import TestCase
from django.urls import NoReverseMatch
from test_views import user_with_perms
from testapp.views import DragonflyViewSet, SightingViewSet

from beam.utils import check_permission, navigation_component_entry, reverse_component


class CheckPermissionsTest(TestCase):
    def test_check_no_permission_implies_true(self):
        self.assertTrue(check_permission(None, user=user_with_perms([]), obj=None))

    def test_check_string_permission(self):
        user = user_with_perms(["testapp.change_dragonfly"])

        self.assertTrue(
            check_permission("testapp.change_dragonfly", user=user, obj=user)
        )
        self.assertFalse(
            check_permission("testapp.view_dragonfly", user=user, obj=user)
        )

    def test_check_callable_permission(self):
        user = user_with_perms([])

        def check(user, obj=None):
            return user == obj

        self.assertTrue(check_permission(check, user=user, obj=user))
        self.assertFalse(check_permission(check, user=user, obj=None))

    def test_check_any_permission_with_no_user_implies_false(self):
        self.assertFalse(check_permission(lambda *args: True, user=None, obj=None))

    def test_navigation_entry_handles_empty(self):
        self.assertEqual(navigation_component_entry(None), None)

    def test_navigation_entry_uses_model_name_as_list_label(self):
        user = user_with_perms(["testapp.view_dragonfly"])
        self.assertEqual(
            navigation_component_entry(DragonflyViewSet().links["list"], user=user),
            ("dragonflys", "/dragonfly/"),
        )

    def test_navigation_entry_handles_other_components(self):
        user = user_with_perms(["testapp.view_sighting"])
        self.assertEqual(
            navigation_component_entry(
                SightingViewSet().links["other_list"], user=user
            ),
            ("Another list is possible", "/sighting/other/"),
        )

    def test_navigation_entry_respects_permissions(self):
        user = user_with_perms(["testapp.view_dragonfly"])
        self.assertEqual(
            navigation_component_entry(DragonflyViewSet().links["list"], user=user),
            ("dragonflys", "/dragonfly/"),
        )
        self.assertEqual(
            navigation_component_entry(SightingViewSet().links["list"], user=user), None
        )

    def test_reverse_component_returns_url(self):
        url = reverse_component(
            component=DragonflyViewSet().links["list"],
            request=None,
            obj=None,
            override_kwargs=None,
        )
        self.assertEqual(url, "/dragonfly/")

    def test_reverse_component_raises_no_reverse_match_if_no_reverse_match(self):
        with self.assertRaises(NoReverseMatch) as e:
            reverse_component(
                component=DragonflyViewSet().links["detail"],
                request=None,
                obj=None,
                override_kwargs=None,
            )
        self.assertIsInstance(e.exception, NoReverseMatch)
        self.assertTrue(
            str(e.exception).startswith(
                "Unable to reverse url to "
                "<Component testapp.views.DragonflyViewSet 'detail'>: "
                "Reverse"
            )
        )

    def test_reverse_component_raises_on_all_exceptions(self):
        component = DragonflyViewSet().links["detail"]
        with mock.patch.object(
            component, "resolve_url", side_effect=ValueError("FAIL")
        ):
            with self.assertRaises(Exception) as e:
                reverse_component(
                    component=component, request=None, obj=None, override_kwargs=None
                )
        self.assertNotIsInstance(e.exception, NoReverseMatch)
        self.assertEqual(
            str(e.exception),
            "Unable to reverse url to <Component testapp.views.DragonflyViewSet 'detail'>: FAIL",
        )



from django.contrib import messages
from django.core.exceptions import ImproperlyConfigured
from django.db import connection, transaction
from django.http import HttpResponseRedirect
from django.utils.encoding import force_str
from django.utils.translation import gettext as _
from django.views import View, generic
from reversion import RevertError, set_comment
from reversion.models import Version

from beam.views import ComponentMixin, InlinesMixin


class _RollBackRevisionView(Exception):
    def __init__(self, response) -> None:
        super().__init__()
        self.response = response


class VersionRestoreView(ComponentMixin, View):
    def post(self, request, *args, **kwargs):
        self.version = Version.objects.get_for_object_reference(
            self.model, kwargs["pk"]
        ).get(pk=kwargs["version_id"])

        # Check that database transactions are supported.
        if not connection.features.uses_savepoints:
            raise ImproperlyConfigured(
                "Cannot use with a database that does not support savepoints."
            )

        # Run the view.
        try:
            with transaction.atomic(using=self.version.db):
                # Revert the revision.
                with self.viewset.create_revision(request):
                    set_comment(_("revert to version {}".format(self.version.pk)))
                    self.version.revision.revert(delete=True)
                    messages.success(request, "Reverted")
                    return HttpResponseRedirect(
                        self.viewset.links["detail"].reverse(
                            obj=self.version.object, request=request
                        )
                    )
        except RevertError as ex:
            messages.error(request, force_str(ex))
            return HttpResponseRedirect(
                self.viewset.links["detail"].reverse(
                    obj=self.version.object, request=request
                )
            )


class VersionDetailView(ComponentMixin, InlinesMixin, generic.DetailView):
    # we don't inherit from DetailView as that would add the ActionMixin
    # we explicitly don't want the ActionMixin in here as that would
    # allow the user to perform the regular actions using an old version
    # as input with rather unpredictable results

    def get_context_data(self, **kwargs):
        kwargs["version"] = self.version
        return super().get_context_data(**kwargs)

    def get(self, request, *args, **kwargs):
        self.version = Version.objects.get_for_object_reference(
            self.model, kwargs["pk"]
        ).get(pk=kwargs["version_id"])

        # Check that database transactions are supported.
        if not connection.features.uses_savepoints:
            raise ImproperlyConfigured(
                "Cannot use with a database that does not support savepoints."
            )

        # Run the view.
        try:
            with transaction.atomic(using=self.version.db):
                self.version.revision.revert(delete=True)
                response = super().get(request, *args, **kwargs)
                if hasattr(response, "render"):
                    response.render()
                raise _RollBackRevisionView(
                    response
                )  # Raise exception to undo the transaction and revision.
        except RevertError as ex:
            messages.error(request, force_str(ex))
            return HttpResponseRedirect(
                self.viewset.links["detail"].reverse(self.version.object)
            )
        except _RollBackRevisionView as ex:
            return ex.response

    def get_template_names(self):
        return ["beam_reversion/version_detail.html"]


class VersionListView(ComponentMixin, generic.DetailView):
    def get_template_names(self):
        return ["beam_reversion/version_list.html"]

    def get_context_data(self, **kwargs):
        kwargs["versions"] = Version.objects.get_for_object(self.object)
        return super().get_context_data(**kwargs)



from contextlib import contextmanager
from typing import Iterable, List

from django.utils.translation import gettext_lazy as _
from reversion import (
    create_revision,
    get_comment,
    is_registered,
    register,
    set_comment,
    set_user,
)

from beam import RelatedInline, ViewSet
from beam.urls import UrlKwargDict
from beam.viewsets import BaseViewSet, Component, DeleteMixin, DetailMixin, UpdateMixin

from .views import VersionDetailView, VersionListView, VersionRestoreView


class VersionRestoreMixin(BaseViewSet):
    version_restore_view_class = VersionRestoreView
    version_restore_url = "<str:pk>/versions/<str:version_id>/restore/"
    version_restore_url_kwargs: UrlKwargDict = {"pk": "pk"}
    version_restore_verbose_name = None
    version_restore_url_name = None
    version_restore_link_layout: List[str] = []
    version_restore_permission = "{app_label}.change_{model_name}"

    def get_component_classes(self):
        return super().get_component_classes() + [("version_restore", Component)]


class VersionDetailMixin(BaseViewSet):
    version_detail_view_class = VersionDetailView
    version_detail_url = "<str:pk>/versions/<str:version_id>/"
    version_detail_url_kwargs: UrlKwargDict = {"pk": "pk"}
    version_detail_verbose_name = _("show version")
    version_detail_url_name = None
    version_detail_link_layout = ["version_list"]
    version_detail_permission = "{app_label}.view_{model_name}"

    detail_link_layout = DetailMixin.detail_link_layout + ["!version_detail"]

    def get_component_classes(self):
        return super().get_component_classes() + [("version_detail", Component)]

    # mirror the detail fields and layout
    @property
    def version_detail_fields(self):
        return getattr(self, "detail_fields", getattr(self, "fields", []))

    @property
    def version_detail_layout(self):
        return getattr(self, "detail_layout", getattr(self, "layout", []))

    @property
    def version_detail_inline_classes(self):
        return getattr(
            self, "detail_inline_classes", getattr(self, "inline_classes", [])
        )


class VersionListMixin(BaseViewSet):
    version_list_view_class = VersionListView
    version_list_url = "<str:pk>/versions/"
    version_list_url_kwargs: UrlKwargDict = {"pk": "pk"}
    version_list_verbose_name = _("history")
    version_list_url_name = None
    version_list_link_layout = ["detail"]
    version_list_permission = "{app_label}.view_{model_name}"

    def get_component_classes(self):
        return super().get_component_classes() + [("version_list", Component)]


class VersionViewSetMixin(VersionDetailMixin, VersionRestoreMixin, VersionListMixin):
    versioned_component_names = ["create", "update", "delete"]
    update_link_layout = UpdateMixin.update_link_layout + [
        "!version_restore",
        "!version_detail",
    ]
    delete_link_layout = DeleteMixin.delete_link_layout + [
        "!version_restore",
        "!version_detail",
    ]

    def __init__(self) -> None:
        super().__init__()
        self.register_model_and_inlines_with_reversion(
            self.model, self._get_version_inline_classes()
        )

    def _get_version_inline_classes(self):
        """
        Get all inline classes that are shown in versioned views.
        """
        return {
            inline_class
            for component in self.components.values()
            for inline_class in component.inline_classes
            if component.name in self.versioned_component_names
        }

    def _register_model_with_parents(self, model, follow=()):
        """
        Register a model with reversion automatically accounting for multi table inheritance.
        """
        if is_registered(model):
            return

        follow = tuple(follow)

        # recursively register parents to support multi table inheritance as per api docs
        # https://django-reversion.readthedocs.io/en/stable/api.html#registration-api
        for parent_model, field in model._meta.concrete_model._meta.parents.items():
            follow += (field.name,)
            self._register_model_with_parents(parent_model, ())

        register(model, follow=follow)

    def register_model_and_inlines_with_reversion(
        self, model, inline_classes: Iterable[RelatedInline]
    ):
        """
        Register the model and all inline models with reversion
        iff they haven't been registered before.

        If you want full control over which models are versioned together,
        register them with reversion manually.
        """
        if is_registered(model):
            return

        # we collect all inline fields so that reversion knows
        # which inline instances to include in revisions.
        inline_fields = []
        for inline_class in inline_classes:
            inline_model = inline_class.model
            inline_model_opts = inline_class.model._meta

            # do not register inlines that have their own viewset, this yields confusing
            # behaviour as they may not be registered on their own
            if (
                inline_model_opts.app_label in self.registry
                and inline_model_opts.model_name
                in self.registry[inline_model_opts.app_label]
            ):
                continue

            # register the inline model
            self._register_model_with_parents(inline_model)

            # if the remote field has an accessor name that
            # we can follow from our model (not is_hidden())
            # we want reversion to follow that accessor when
            # creating a revision for our model
            if not inline_model._meta.get_field(
                inline_class.foreign_key_field
            ).remote_field.is_hidden():
                field = inline_model._meta.get_field(inline_class.foreign_key_field)
                inline_fields.append(field.remote_field.get_accessor_name())

        self._register_model_with_parents(model, inline_fields)

    @contextmanager
    def create_revision(self, request):
        with create_revision():
            if request.user.is_authenticated:
                set_user(request.user)
            yield

    def _set_revision_comment(self, component, request, *args, **kwargs):
        """
        Sets the revision comment according to component and parameters.
        """
        if not get_comment():
            comment = component.verbose_name
            set_comment(comment)

    def _get_view(self, component):
        """
        Ensure that the views in `self.versioned_component_names`
        are wrapped with the create_revision context_manager
        """
        view = super()._get_view(component)
        if component.name not in self.versioned_component_names:
            return view

        def wrapped_view(request, *args, **kwargs):
            with self.create_revision(request):
                self._set_revision_comment(component, request, *args, **kwargs)
                return view(request, *args, **kwargs)

        return wrapped_view


class VersionViewSet(VersionViewSetMixin, ViewSet):
    pass



from collections import OrderedDict
from functools import wraps
from logging import getLogger
from typing import Any, Dict, Iterable, List, Mapping, Optional, Sequence, Tuple, Type

import django_filters
from django.db.models import Model, QuerySet
from django.forms import Form, ModelForm
from django.urls import path
from django.utils.functional import cached_property
from django.utils.translation import gettext_lazy as _
from django.views import View

from beam.registry import ViewsetMetaClass, default_registry

from .actions import Action
from .components import BaseComponent, Component, FormComponent, ListComponent
from .inlines import RelatedInline
from .types import LayoutType
from .urls import UrlKwargDict
from .views import CreateView, DeleteView, DetailView, ListView, UpdateView

logger = getLogger(__name__)


undefined = (
    object()
)  # sentinel value for attributes not defined or overwritten on the viewset


class BaseViewSet(metaclass=ViewsetMetaClass):
    registry = default_registry

    model: Model
    fields: List[str]
    layout: LayoutType
    queryset: QuerySet
    inline_classes: List[Type[RelatedInline]] = []
    form_class: Form
    link_layout: List[str]
    url_namespace: str = ""

    # we default to change_ because it is a safe default
    permission = "{app_label}.change_{model_name}"

    _component_classes: Sequence[Tuple[str, Type[Component]]] = []

    def get_component_classes(self) -> Sequence[Tuple[str, Type[Component]]]:
        return self._component_classes

    def _get_components(self) -> Dict[str, Component]:
        components: Dict[str, Component] = OrderedDict()
        for name, component in self.get_component_classes():
            kwargs = self._resolve_component_kwargs(name, component.get_arguments())
            components[name] = component(**kwargs)
        return components

    def _resolve_component_kwargs(self, component_name, arguments: Iterable[str]):
        component_kwargs = {}

        specific_prefix = "{}_".format(component_name)
        for name in arguments:
            if name in ["name", "viewset"]:  # those are set below
                continue
            specific_value = getattr(self, specific_prefix + name, undefined)

            if specific_value is not undefined:
                component_kwargs[name] = specific_value
                continue

            value = getattr(self, name, undefined)
            if value is not undefined:
                component_kwargs[name] = value
                continue

        component_kwargs["name"] = component_name
        component_kwargs["viewset"] = self

        return component_kwargs

    @cached_property
    def components(self) -> Dict[str, Component]:
        return self._get_components()

    @property
    def links(self) -> Dict[str, BaseComponent]:
        """A list of components that can be linked from within the ui"""
        return self.components

    def _get_view(self, component: Component):
        # FIXME handle function based views?
        view_class = component.view_class
        view_kwargs: Dict[str, Any] = {}

        view = view_class.as_view(**view_kwargs)
        if hasattr(view_class, "viewset"):

            @wraps(view)
            def wrapped_view(request, *args, **kwargs):
                view.view_initkwargs["viewset"] = self
                view.view_initkwargs["component"] = component
                return view(request, *args, **kwargs)

            return wrapped_view
        else:
            return view

    def _get_url_pattern(self, component):
        view = self._get_view(component)
        url = component.url
        url_name = component.url_name
        return path(url, view, name=url_name)

    def get_urls(self):
        urlpatterns = []
        components = list(self.components.values())

        # move patterns without kwargs to the front
        for component in components:
            if not getattr(component, "url_kwargs", None):
                urlpatterns.append(self._get_url_pattern(component))

        for component in components:
            if getattr(component, "url_kwargs", None):
                urlpatterns.append(self._get_url_pattern(component))

        return urlpatterns


class ListMixin(BaseViewSet):
    list_component = ListComponent
    list_view_class = ListView
    list_url = ""
    list_url_name: str
    list_url_kwargs: UrlKwargDict = {}
    list_verbose_name = _("list")

    list_sort_fields: List[str]
    list_sort_fields_columns: Mapping[str, str]
    list_search_fields: List[str] = []
    list_paginate_by = 25
    list_item_link_layout = ["update", "detail"]

    list_model: Model
    list_fields: List[str]
    list_layout: LayoutType
    list_queryset: QuerySet
    list_inline_classes: List[Type[RelatedInline]]
    list_form_class: ModelForm
    list_permission = "{app_label}.view_{model_name}"
    list_filterset_fields: List[str] = []
    list_filterset_class: Optional[Type[django_filters.FilterSet]] = None
    list_action_classes: List[Type[Action]] = []
    list_link_layout = ["create"]


class CreateMixin(BaseViewSet):
    create_component = FormComponent
    create_view_class = CreateView
    create_url = "create/"
    create_url_kwargs: UrlKwargDict = {}
    create_url_name: str
    create_verbose_name = _("create")

    create_model: Model
    create_fields: List[str]
    create_layout: LayoutType
    create_queryset: QuerySet
    create_inline_classes: List[Type[RelatedInline]]
    create_form_class: ModelForm
    create_link_layout = ["list"]
    create_permission = "{app_label}.add_{model_name}"


class UpdateMixin(BaseViewSet):
    update_component = FormComponent
    update_view_class = UpdateView
    update_url = "<str:pk>/update/"
    update_url_kwargs: UrlKwargDict = {"pk": "pk"}
    update_url_name: str
    update_verbose_name = _("update")

    update_model: Model
    update_fields: List[str]
    update_layout: LayoutType
    update_queryset: QuerySet
    update_inline_classes: List[Type[RelatedInline]]
    update_form_class: ModelForm
    update_link_layout = ["!create", "!update", "list", "...", "detail"]
    update_permission = "{app_label}.change_{model_name}"


class DetailMixin(BaseViewSet):
    detail_component = Component
    detail_view_class: View = DetailView
    detail_url: str = "<str:pk>/"
    detail_url_name: str
    detail_url_kwargs: UrlKwargDict = {"pk": "pk"}
    detail_verbose_name = _("detail")

    detail_model: Model
    detail_fields: List[str]
    detail_layout: LayoutType
    detail_queryset: QuerySet
    detail_inline_classes: List[Type[RelatedInline]]
    detail_link_layout = ["!detail", "...", "update"]
    detail_permission = "{app_label}.view_{model_name}"


class DeleteMixin(BaseViewSet):
    delete_component = Component
    delete_view_class = DeleteView
    delete_url = "<str:pk>/delete/"
    delete_url_name: str
    delete_url_kwargs: UrlKwargDict = {"pk": "pk"}
    delete_verbose_name = _("delete")

    delete_model: Model
    delete_fields: List[str]
    delete_layout: LayoutType
    delete_queryset: QuerySet
    delete_inline_classes: List[Type[RelatedInline]]
    delete_link_layout = ["!delete", "..."]
    delete_permission = "{app_label}.delete_{model_name}"


class ViewSet(
    DeleteMixin, UpdateMixin, DetailMixin, CreateMixin, ListMixin, BaseViewSet
):
    pass


__all__ = [
    "LayoutType",
    "undefined",
    "BaseViewSet",
    "ListMixin",
    "CreateView",
    "UpdateMixin",
    "DetailMixin",
    "DeleteMixin",
    "ViewSet",
]



from typing import List, Optional, Type

from django.apps import apps
from django.contrib import messages
from django.contrib.admin.utils import NestedObjects
from django.core.exceptions import FieldDoesNotExist, PermissionDenied
from django.db import router
from django.forms import all_valid
from django.http import HttpResponse, HttpResponseForbidden
from django.shortcuts import redirect
from django.utils.html import escape
from django.utils.translation import gettext as _
from django.views import generic
from django.views.generic.base import ContextMixin, TemplateView
from django_filters.filterset import filterset_factory
from extra_views import SearchableListMixin

from beam.registry import default_registry, register

from .actions import Action
from .components import Component, ListComponent
from .inlines import RelatedInline


class ComponentMixin(ContextMixin):
    component: Optional[Component] = None
    viewset = None

    def get_template_names(self):
        template_names = super().get_template_names()
        opts = self.model._meta
        component_template_name = (
            f"{opts.app_label}/{opts.model_name}_{self.component.name}.html"
        )
        insert_index = len(template_names)

        # we check if a generic template via template_name_suffix is contained in the
        # template names, if so we want to insert the component template before that
        # so that e.g. foo/bar_create.html takes precedence over foo/bar_form.html
        template_name_suffix = getattr(self, "template_name_suffix", None)
        if template_name_suffix:
            suffix_template_name = (
                f"{opts.app_label}/{opts.model_name}{template_name_suffix}.html"
            )
            try:
                insert_index = template_names.index(suffix_template_name)
            except ValueError:
                pass

        if component_template_name not in template_names:
            template_names.insert(insert_index, component_template_name)

        return template_names

    def get_context_data(self, **kwargs):
        context = super().get_context_data(**kwargs)

        context["viewset"] = self.viewset
        context["component"] = self.component
        context["popup"] = self.request.GET.get("_popup")

        return context

    @property
    def model(self):
        return self.component.model

    def get_queryset(self):
        return self.component.queryset

    def get_form_class(self):
        if getattr(self.component, "form_class", None):
            return self.component.form_class
        return super().get_form_class()

    @property
    def fields(self):
        if self.component.fields:
            return self.component.fields
        return super().fields

    def get_inline_classes(self):
        return self.component.inline_classes

    def has_perm(self):
        try:
            obj = self.get_object()
        except AttributeError:
            obj = None
        return self.component.has_perm(self.request.user, obj)

    def handle_no_permission(self):
        if self.request.user.is_authenticated:
            raise PermissionDenied("You shall not pass")

        # Inner import to prevent django.contrib.admin (app) from
        # importing django.contrib.auth.models.User (unrelated model).
        from django.contrib.auth.views import redirect_to_login

        return redirect_to_login(self.request.get_full_path())

    def dispatch(self, request, *args, **kwargs):
        if not self.has_perm():
            return self.handle_no_permission()

        return super().dispatch(request, *args, **kwargs)


class InlinesMixin(ContextMixin):
    inline_classes: List[Type[RelatedInline]] = []

    def get_inline_classes(self):
        return self.inline_classes

    def get_inlines(self, object=None):
        inlines = []
        for inline_class in self.get_inline_classes():
            inlines.append(
                inline_class(
                    parent_instance=object if object is not None else self.object,
                    parent_model=self.model,
                    request=self.request,
                )
            )
        return inlines

    def get_context_data(self, **kwargs):
        if "inlines" not in kwargs:
            kwargs["inlines"] = self.get_inlines()
        return super().get_context_data(**kwargs)


class CreateWithInlinesMixin(InlinesMixin):
    def post(self, request, *args, **kwargs):
        self.object = None
        form_class = self.get_form_class()
        form = self.get_form(form_class)

        if form.is_valid():
            # we have to make sure that the same instance is used for form and inlines
            inlines = self.get_inlines(object=form.save(commit=False))
        else:
            inlines = self.get_inlines()

        if all_valid(inline.formset for inline in inlines) and form.is_valid():
            return self.form_valid(form, inlines)

        return self.form_invalid(form, inlines)

    def form_valid(self, form, inlines):
        self.object = form.save()
        for inline in inlines:
            inline.formset.save()
        return redirect(self.get_success_url())

    def form_invalid(self, form, inlines):
        return self.render_to_response(
            self.get_context_data(form=form, inlines=inlines)
        )


class UpdateWithInlinesMixin(InlinesMixin):
    def post(self, request, *args, **kwargs):
        self.object = self.get_object()
        form_class = self.get_form_class()
        form = self.get_form(form_class)

        inlines = self.get_inlines()

        if form.is_valid() and all_valid(inline.formset for inline in inlines):
            return self.form_valid(form, inlines)

        return self.form_invalid(form, inlines)

    def form_valid(self, form, inlines):
        self.object = form.save()
        for inline in inlines:
            inline.formset.save()
        return redirect(self.get_success_url())

    def form_invalid(self, form, inlines):
        return self.render_to_response(
            self.get_context_data(form=form, inlines=inlines)
        )


class CreateView(ComponentMixin, CreateWithInlinesMixin, generic.CreateView):
    def get_template_names(self):
        return super().get_template_names() + ["beam/create.html"]

    def get_success_url(self):
        return self.viewset.links["detail"].reverse(
            obj=self.object, request=self.request
        )

    def get_success_message(self):
        return _('The {model} "{name}" was added successfully.').format(
            model=self.model._meta.verbose_name,
            name=str(self.object),
        )

    def form_valid(self, form, inlines):
        response = super().form_valid(form, inlines)
        success_message = self.get_success_message()
        if success_message:
            messages.success(self.request, success_message)
        if self.request.GET.get("_popup"):
            return self.popup_response()
        return response

    def popup_response(self):
        return HttpResponse(
            "<script>"
            "window.opener.postMessage("
            '{{id: "{id}", result: "created", source: "{source}", text: "{text}"}}, '
            "document.origin"
            ");"
            "window.close()"
            "</script>".format(
                id=escape(self.object.pk),
                source=escape(self.request.GET["_popup"]),
                text=escape(str(self.object)),
            )
        )


class UpdateView(ComponentMixin, UpdateWithInlinesMixin, generic.UpdateView):
    def get_template_names(self):
        return super().get_template_names() + ["beam/update.html"]

    def get_success_message(self):
        return _('The {model} "{name}" was changed successfully.').format(
            model=self.model._meta.verbose_name,
            name=str(self.object),
        )

    def form_valid(self, form, inlines):
        response = super().form_valid(form, inlines)
        success_message = self.get_success_message()
        if success_message:
            messages.success(self.request, success_message)
        return response

    def get_success_url(self):
        if self.request.POST.get("submit", None) == "save_and_continue_editing":
            return self.request.get_full_path()
        return self.viewset.links["detail"].reverse(
            obj=self.object, request=self.request
        )


class SortableListMixin(ComponentMixin):
    sort_param = "o"
    sort_separator = ","

    def get_sort_fields(self):
        if self.component.list_sort_fields is None:
            return [
                # cast to string to support virtual fields
                str(field)
                for field in self.component.fields
                if self.get_sort_column_for_field(str(field))
            ]

        for field in self.component.list_sort_fields:
            if self.get_sort_column_for_field(field) is None:
                raise Exception(
                    "Unable to determine sort column for explicit sort field {} on {}".format(
                        field, self.viewset
                    )
                )

        return self.component.list_sort_fields

    def get_sort_fields_columns(self):
        return self.component.list_sort_fields_columns or {}

    def get_sort_column_for_field(self, field_name):
        explicit = self.get_sort_fields_columns()
        if field_name in explicit:
            return explicit[field_name]

        try:
            field = self.model._meta.get_field(field_name)
            return field.name
        except FieldDoesNotExist:
            return None

    def get_sort_fields_from_request(self) -> List[str]:
        fields = []
        sort_fields = set(self.get_sort_fields())
        for field in self.request.GET.get(self.sort_param, "").split(
            self.sort_separator
        ):
            if field.startswith("-"):
                sort_field = field[1:]
            else:
                sort_field = field

            if sort_field in sort_fields:
                fields.append(field)
        return fields

    def get_sort_columns(self, fields):
        columns = []
        for field in fields:
            if field.startswith("-"):
                descending = True
                field = field[1:]
            else:
                descending = False

            column = self.get_sort_column_for_field(field)
            if not column:
                continue

            columns.append("-" + column if descending else column)
        return columns

    def sort_queryset(self, qs):
        current_sort_fields = self.get_sort_fields_from_request()
        current_sort_columns = self.get_sort_columns(current_sort_fields)
        if current_sort_columns:
            qs = qs.order_by(*current_sort_columns)
        return qs

    def get_queryset(self):
        qs = super().get_queryset()
        return self.sort_queryset(qs)

    def get_context_data(self, **kwargs):
        context = super().get_context_data(**kwargs)
        context["sortable_fields"] = set(self.get_sort_fields())
        context["sorted_fields"] = self.get_sort_fields_from_request()
        return context


class FiltersetMixin(ComponentMixin):
    filterset_class = None
    filterset_fields = None
    filterset = None

    def get_filterset_fields(self):
        if self.filterset_fields is not None:
            return self.filterset_fields
        return self.component.list_filterset_fields

    def get_filterset_class(self):
        if self.filterset_class:
            return self.filterset_class
        if self.component.list_filterset_class:
            return self.component.list_filterset_class
        elif self.component.list_filterset_fields:
            return filterset_factory(
                model=self.model, fields=self.get_filterset_fields()
            )
        return None

    def get_filterset_kwargs(self):
        """
        Returns the keyword arguments for instantiating the filterset.
        """
        kwargs = {
            "data": self.request.GET or None,
            "request": self.request,
            "prefix": "filter",
            "queryset": self.component.queryset,
        }
        return kwargs

    def get_filterset(self):
        filterset_class = self.get_filterset_class()
        if not filterset_class:
            return None
        return filterset_class(**self.get_filterset_kwargs())

    def get_queryset(self):
        qs = super().get_queryset()
        if self.filterset and self.filterset.is_bound and self.filterset.is_valid():
            qs = self.filterset.filter_queryset(qs)
        return qs

    def dispatch(self, request, *args, **kwargs):
        self.filterset = self.get_filterset()
        return super().dispatch(request, *args, **kwargs)

    def get_context_data(self, **kwargs):
        context = super().get_context_data(**kwargs)
        context["filterset"] = self.filterset
        return context


class InlineActionMixin(InlinesMixin):
    def get_action_qs(self, inline):
        ids = self.request.POST.getlist("_action_select[]")
        select_across = self.request.POST.get("_action_select_across") == "all"

        objects = inline.get_queryset()
        if not select_across:
            objects = objects.filter(pk__in=ids)

        if not select_across and len(objects) != len(ids):
            messages.error(
                self.request,
                _(
                    "There was an error finding the objects you selected. "
                    "This could be caused by another user changing them concurrently. "
                    "Please try again."
                ),
            )
            return objects.none()

        return objects

    def get_action(self):
        for inline in self.get_inlines(self.get_object()):
            action = inline.get_action()
            if action and action.is_bound:
                return inline, action
        return None, None

    def handle_action(self, inline, action):
        form = action.get_form()

        if form and not form.is_valid():
            return None

        result: Optional[HttpResponse] = action.apply(
            queryset=self.get_action_qs(inline)
        )
        success_message: str = action.get_success_message()

        if success_message:
            messages.success(self.request, success_message)

        if result:
            return result

        return redirect(self.request.get_full_path())

    def post(self, request, *args, **kwargs):
        inline, action = self.get_action()
        if action:
            response = self.handle_action(inline, action)
            if response:
                return response
        return self.get(request, *args, **kwargs)


class ListActionsMixin(ComponentMixin):
    component: ListComponent

    def get_context_data(self, **kwargs):
        context = super().get_context_data(**kwargs)
        context["actions"] = self.actions
        return context

    def get_action_qs(self):
        ids = self.request.POST.getlist("_action_select[]")
        select_across = self.request.POST.get("_action_select_across") == "all"

        objects = self.get_queryset()
        if not select_across:
            objects = objects.filter(pk__in=ids)

        if not select_across and len(objects) != len(ids):
            messages.error(
                self.request,
                _(
                    "There was an error finding the objects you selected. "
                    "This could be caused by another user changing them concurrently. "
                    "Please try again."
                ),
            )
            return objects.none()

        return objects

    def get_actions(self):
        selected_action = self.request.POST.get("_action_choice")
        actions = []
        action_class: Type[Action]
        for index, action_class in enumerate(self.component.list_actions_classes):
            action_id = "{}-{}".format(index, action_class.name)
            action = action_class(
                data=self.request.POST if action_id == selected_action else None,
                model=self.model,
                id=action_id,
                request=self.request,
            )
            if action.has_perm(self.request.user):
                actions.append(action)
        return actions

    def get_action(self):
        for action in self.actions:
            if action.is_bound:
                return action
        return None

    def handle_action(self, action):
        form = action.get_form()

        if form and not form.is_valid():
            return None

        result: Optional[HttpResponse] = action.apply(queryset=self.get_action_qs())
        success_message: str = action.get_success_message()

        if success_message:
            messages.success(self.request, success_message)

        if result:
            return result

        return redirect(self.request.get_full_path())

    def dispatch(self, request, *args, **kwargs):
        self.actions = self.get_actions()
        return super().dispatch(request, *args, **kwargs)

    def post(self, request, *args, **kwargs):
        action = self.get_action()
        if action:
            response = self.handle_action(action)
            if response:
                return response
        return self.get(request, *args, **kwargs)


class ListView(
    ListActionsMixin,
    FiltersetMixin,
    SearchableListMixin,
    SortableListMixin,
    ComponentMixin,
    generic.ListView,
):
    @property
    def search_fields(self):
        return self.component.list_search_fields

    def get_paginate_by(self, queryset):
        return self.component.list_paginate_by

    def get_search_query(self):
        if not self.search_fields:
            return ""
        return super().get_search_query()

    def get_template_names(self):
        return super().get_template_names() + ["beam/list.html"]

    def get_context_data(self, **kwargs):
        context = super().get_context_data(**kwargs)
        context["search_query"] = self.get_search_query()
        context["list_item_link_layout"] = self.component.list_item_link_layout
        return context


class DetailView(InlineActionMixin, ComponentMixin, InlinesMixin, generic.DetailView):
    def get_template_names(self):
        return super().get_template_names() + ["beam/detail.html"]


class DeleteView(ComponentMixin, InlinesMixin, generic.DeleteView):
    def get_template_names(self):
        return super().get_template_names() + ["beam/delete.html"]

    def get_success_url(self):
        return self.viewset.links["list"].reverse(request=self.request)

    def get_success_message(self):
        return _('The {model} "{name}" was deleted successfully.').format(
            model=self.model._meta.verbose_name, name=str(self.object)
        )

    def post(self, request, *args, **kwargs):
        self.object = self.get_object()
        nested, protected = self.get_nested_objects(self.object)
        if protected:
            return HttpResponseForbidden()

        success_message = self.get_success_message()

        response = self.delete(request, *args, **kwargs)

        if success_message:
            messages.success(request, success_message)

        return response

    @classmethod
    def get_nested_objects(cls, obj):
        using = router.db_for_write(cls.model)
        collector = NestedObjects(using=using)
        collector.collect([obj])
        nested = collector.nested(cls._format_obj)
        return nested, list(map(cls._format_obj, collector.protected))

    @staticmethod
    def _format_obj(obj):
        return '%s "%s"' % (obj._meta.verbose_name, str(obj))

    def get_context_data(self, **kwargs):
        context = super(DeleteView, self).get_context_data(**kwargs)
        nested, protected = self.get_nested_objects(self.get_object())
        context.update(
            {
                "object": self.object,
                "object_name": self._format_obj(self.object),
                "nested_objects": nested,
                "protected_objects": protected,
            }
        )
        return context


class DashboardView(TemplateView):
    template_name = "beam/dashboard.html"

    viewsets = None
    registry = default_registry

    def build_registry(self, viewsets):
        registry = {}
        for viewset in viewsets:
            register(registry, viewset)
        return registry

    def get_registry(self):
        if self.viewsets:
            return self.build_registry(self.viewsets)
        else:
            return self.registry

    def get_context_data(self, **kwargs):
        context = super().get_context_data(**kwargs)

        grouped = []
        for app_label, viewsets_dict in self.get_registry().items():
            viewsets = []
            for viewset in viewsets_dict.values():
                links = []
                for name in "list", "create":
                    link = viewset().links.get(name)
                    if link and link.has_perm(
                        user=self.request.user, obj=None, request=self.request
                    ):
                        links.append(link)

                if links:
                    viewsets.append((viewset, links))

            if not viewsets:
                continue

            group = {
                "app_label": app_label,
                "app_config": apps.get_app_config(app_label),
                "viewsets": viewsets,
            }
            grouped.append(group)
        context["grouped_by_app"] = grouped
        return context



from django.contrib.auth import get_user_model
from django.contrib.auth.hashers import check_password
from django.contrib.auth.models import Group, Permission
from django.test.utils import override_settings
from django.urls import include, path, reverse
from django_webtest import WebTest
from test_views import user_with_perms

urlpatterns = [
    path("auth/", include("beam.contrib.auth.urls")),
]


@override_settings(ROOT_URLCONF=__name__)
class UserViewTest(WebTest):
    def test_create_user(self):
        create_page = self.app.get(
            reverse("auth_user_create"),
            user=user_with_perms(["auth.add_user", "auth.view_user"]),
        )
        form = create_page.form
        form["username"] = "a.user"
        form["password1"] = "v3rys3cur3"
        form["password2"] = "v3rys3cur3"

        response = form.submit()

        self.assertContains(response.follow(), "a.user")

        user = get_user_model().objects.get(username="a.user")

        self.assertTrue(check_password("v3rys3cur3", user.password))

    def test_update_user(self):
        user = user_with_perms(["auth.delete_group"], username="other.user")
        one_group = Group.objects.create(name="One group")
        other_group = Group.objects.create(name="Other group")
        user.groups.add(one_group)

        update_page = self.app.get(
            reverse("auth_user_update", kwargs={"pk": user.pk}),
            user=user_with_perms(
                ["auth.add_user", "auth.view_user", "auth.change_user"]
            ),
        )

        form = update_page.form

        form["email"] = "a.user@example.com"
        form["groups"] = [other_group.pk]
        form["user_permissions"] = [
            Permission.objects.get(codename="add_group").pk,
            Permission.objects.get(codename="view_group").pk,
        ]

        response = form.submit()

        self.assertContains(response.follow(), "a.user")

        user.refresh_from_db()

        self.assertEqual(user.email, "a.user@example.com")

        self.assertSequenceEqual(user.groups.all(), [other_group])

        self.assertSetEqual(
            set(user.get_all_permissions()), {"auth.add_group", "auth.view_group"}
        )

    def test_user_detail(self):
        user = user_with_perms(["auth.view_group"], username="other.user")
        one_group = Group.objects.create(name="One group")
        Group.objects.create(name="Other group")
        user.groups.add(one_group)

        detail_page = self.app.get(
            reverse("auth_user_detail", kwargs={"pk": user.pk}),
            user=user_with_perms(
                ["auth.add_user", "auth.view_user", "auth.change_user"]
            ),
        )

        self.assertContains(detail_page, "other.user")
        self.assertContains(detail_page, "One group")
        self.assertNotContains(detail_page, "Other group")
        self.assertContains(detail_page, "view_group")
        self.assertNotContains(detail_page, "change_group")

    def test_user_list(self):
        one_user = user_with_perms([], username="one.user")
        other_user = user_with_perms([], username="other.user")
        list_page = self.app.get(
            reverse("auth_user_list"),
            user=user_with_perms(
                ["auth.add_user", "auth.view_user", "auth.change_user"],
                username="admin.user",
            ),
        )
        self.assertContains(list_page, str(one_user))
        self.assertContains(list_page, str(other_user))

    def test_user_list_search(self):
        user_with_perms([], username="one.user")
        user_with_perms([], username="other.user")

        list_page = self.app.get(
            reverse("auth_user_list"),
            user=user_with_perms(
                ["auth.add_user", "auth.view_user", "auth.change_user"],
                username="admin.user",
            ),
        )

        list_page.forms["search-form"]["q"] = "one.u"
        search_response = list_page.forms["search-form"].submit()

        self.assertContains(search_response, "one.user")
        self.assertNotContains(search_response, "other.user")

    def test_user_list_filter(self):
        one_user = user_with_perms([], username="one.user")
        other_user = user_with_perms([], username="other.user")

        other_user.is_superuser = True
        other_user.save()

        one_group = Group.objects.create(name="One group")
        one_group.user_set.add(one_user)
        other_group = Group.objects.create(name="Other group")
        other_group.user_set.add(other_user)

        list_page = self.app.get(
            reverse("auth_user_list"),
            user=user_with_perms(
                ["auth.add_user", "auth.view_user", "auth.change_user"],
            ),
        )

        filter_form = list_page.forms["filter-form"]
        filter_form["filter-groups"] = [one_group.pk]
        filter_response = filter_form.submit()

        self.assertContains(filter_response, "one.user")
        self.assertNotContains(filter_response, "other.user")

        filter_form = list_page.forms["filter-form"]
        filter_form["filter-groups"] = []
        filter_form["filter-is_superuser"] = "true"
        filter_response = filter_form.submit()

        self.assertContains(filter_response, "other.user")
        self.assertNotContains(filter_response, "one.user")

    def test_user_delete(self):
        one_user = user_with_perms(["auth.view_group"], username="one.user")
        one_group = Group.objects.create(name="One group")
        one_group.user_set.add(one_user)

        delete_page = self.app.get(
            reverse("auth_user_delete", kwargs={"pk": one_user.pk}),
            user=user_with_perms(
                ["auth.delete_user", "auth.view_user"],
            ),
        )

        self.assertContains(delete_page, "user-group relationship")
        self.assertContains(delete_page, "user-permission relationship")

        delete_response = delete_page.form.submit()

        self.assertRedirects(delete_response, reverse("auth_user_list"))

        self.assertFalse(get_user_model().objects.filter(pk=one_user.pk).exists())

    def test_user_create_requires_permission(self):
        self.app.get(
            reverse("auth_user_create"),
            user=user_with_perms([], username="no-perms"),
            status=403,
        )
        self.app.get(
            reverse("auth_user_create"),
            user=user_with_perms(["auth.add_user"], username="perms"),
            status=200,
        )

    def test_user_list_requires_permission(self):
        self.app.get(
            reverse("auth_user_list"),
            user=user_with_perms([], username="no-perms"),
            status=403,
        )
        self.app.get(
            reverse("auth_user_list"),
            user=user_with_perms(["auth.view_user"], username="perms"),
            status=200,
        )

    def test_user_detail_requires_permission(self):
        user = user_with_perms([], username="no-perms")
        self.app.get(
            reverse("auth_user_detail", kwargs={"pk": user.pk}),
            user=user,
            status=403,
        )
        self.app.get(
            reverse("auth_user_detail", kwargs={"pk": user.pk}),
            user=user_with_perms(["auth.view_user"], username="perms"),
            status=200,
        )

    def test_user_update_requires_permission(self):
        user = user_with_perms([], username="no-perms")
        self.app.get(
            reverse("auth_user_update", kwargs={"pk": user.pk}),
            user=user,
            status=403,
        )
        self.app.get(
            reverse("auth_user_update", kwargs={"pk": user.pk}),
            user=user_with_perms(["auth.change_user"], username="perms"),
            status=200,
        )

    def test_user_delete_requires_permission(self):
        user = user_with_perms([], username="no-perms")
        self.app.get(
            reverse("auth_user_delete", kwargs={"pk": user.pk}),
            user=user,
            status=403,
        )
        self.app.get(
            reverse("auth_user_delete", kwargs={"pk": user.pk}),
            user=user_with_perms(["auth.delete_user"], username="perms"),
            status=200,
        )


@override_settings(ROOT_URLCONF=__name__)
class GroupViewTest(WebTest):
    def test_group_create_requires_permission(self):
        self.app.get(
            reverse("auth_group_create"),
            user=user_with_perms([], username="no-perms"),
            status=403,
        )
        self.app.get(
            reverse("auth_group_create"),
            user=user_with_perms(["auth.add_group"], username="perms"),
            status=200,
        )

    def test_group_list_requires_permission(self):
        self.app.get(
            reverse("auth_group_list"),
            user=user_with_perms([], username="no-perms"),
            status=403,
        )
        self.app.get(
            reverse("auth_group_list"),
            user=user_with_perms(["auth.view_group"], username="perms"),
            status=200,
        )

    def test_group_detail_requires_permission(self):
        group = Group.objects.create()
        self.app.get(
            reverse("auth_group_detail", kwargs={"pk": group.pk}),
            user=user_with_perms([], username="no-perms"),
            status=403,
        )
        self.app.get(
            reverse("auth_group_detail", kwargs={"pk": group.pk}),
            user=user_with_perms(["auth.view_group"], username="perms"),
            status=200,
        )

    def test_group_update_requires_permission(self):
        group = Group.objects.create()
        self.app.get(
            reverse("auth_group_update", kwargs={"pk": group.pk}),
            user=user_with_perms([], username="no-perms"),
            status=403,
        )
        self.app.get(
            reverse("auth_group_update", kwargs={"pk": group.pk}),
            user=user_with_perms(["auth.change_group"], username="perms"),
            status=200,
        )

    def test_group_delete_requires_permission(self):
        group = Group.objects.create()
        self.app.get(
            reverse("auth_group_delete", kwargs={"pk": group.pk}),
            user=user_with_perms([], username="no-perms"),
            status=403,
        )
        self.app.get(
            reverse("auth_group_delete", kwargs={"pk": group.pk}),
            user=user_with_perms(["auth.delete_group"], username="perms"),
            status=200,
        )

    def test_create_group(self):
        create_page = self.app.get(
            reverse("auth_group_create"),
            user=user_with_perms(["auth.add_group", "auth.view_group"]),
        )
        form = create_page.form
        form["name"] = "A group"
        form["permissions"] = [Permission.objects.get(codename="view_user").pk]

        response = form.submit()

        self.assertContains(response.follow(), "A group")
        self.assertContains(response.follow(), "view_user")

        group = Group.objects.get(name="A group")

        self.assertTrue(group.permissions.filter(codename="view_user").exists())

    def test_update_group(self):
        group = Group.objects.create(name="One group")
        group.permissions.add(Permission.objects.get(codename="view_user"))

        update_page = self.app.get(
            reverse("auth_group_update", kwargs={"pk": group.pk}),
            user=user_with_perms(["auth.change_group", "auth.view_group"]),
        )

        form = update_page.form

        form["name"] = "1 group"

        form["permissions"] = [
            Permission.objects.get(codename="view_group").pk,
            Permission.objects.get(codename="change_group").pk,
        ]
        response = form.submit()

        self.assertContains(response.follow(), "1 group")

        group.refresh_from_db()

        self.assertEqual(group.name, "1 group")

        self.assertSetEqual(
            set(group.permissions.values_list("codename", flat=True)),
            {"change_group", "view_group"},
        )

    def test_group_detail(self):
        user = user_with_perms(["auth.view_group"], username="a.user")
        group = Group.objects.create(name="One group")
        group.permissions.add(Permission.objects.get(codename="view_user"))
        group.user_set.add(user)

        detail_page = self.app.get(
            reverse("auth_group_detail", kwargs={"pk": group.pk}),
            user=user,
        )

        self.assertContains(detail_page, "One group")
        self.assertContains(detail_page, "view_user")
        self.assertContains(detail_page, "a.user")

    def test_group_delete(self):
        user = user_with_perms(
            ["auth.view_group", "auth.delete_group"], username="a.user"
        )
        group = Group.objects.create(name="One group")
        group.permissions.add(Permission.objects.get(codename="view_user"))
        group.user_set.add(user)

        delete_page = self.app.get(
            reverse("auth_group_delete", kwargs={"pk": group.pk}),
            user=user,
        )

        self.assertContains(delete_page, "user-group relationship")
        self.assertContains(delete_page, "group-permission relationship")

        delete_response = delete_page.form.submit()

        self.assertRedirects(delete_response, reverse("auth_group_list"))

        self.assertFalse(Group.objects.filter(pk=group.pk).exists())

    def test_group_list(self):
        Group.objects.create(name="One group")
        Group.objects.create(name="Other group")
        list_page = self.app.get(
            reverse("auth_group_list"),
            user=(user_with_perms(["auth.view_group"], username="a.user")),
        )
        self.assertContains(list_page, "One group")
        self.assertContains(list_page, "Other group")

    def test_group_list_search(self):
        Group.objects.create(name="One group")
        Group.objects.create(name="Other group")

        list_page = self.app.get(
            reverse("auth_group_list"),
            user=(user_with_perms(["auth.view_group"], username="a.user")),
        )

        search_from = list_page.forms["search-form"]
        search_from["q"] = "Other"
        search_results = search_from.submit()

        self.assertNotContains(search_results, "One group")
        self.assertContains(search_results, "Other group")



from django.template import RequestContext
from django.test import RequestFactory, TestCase
from django.urls import NoReverseMatch
from test_views import user_with_perms
from testapp.models import Dragonfly
from testapp.views import DragonflyViewSet

from beam.templatetags.beam_tags import (
    _add_params_to_url_if_new,
    get_apps_for_navigation,
    get_visible_links,
    preserve_query_string,
)


class TagsTest(TestCase):
    def test_preserve_query_string_preserves_query_string(self):
        request = RequestFactory().get("/", data={"a": "b"})
        query = preserve_query_string({"request": request}, page="1")
        self.assertEqual(query, "?a=b&page=1")

    def test_preserve_query_string_works_with_mulitple_arguments(self):
        request = RequestFactory().get("/", data={"a": ["b", "c"]})
        query = preserve_query_string({"request": request}, page="1")
        self.assertEqual(query, "?a=b&a=c&page=1")

    def test_add_url_params_works(self):
        url = "https://example.com"
        self.assertEqual(
            _add_params_to_url_if_new(url, {"foo": "bar"}),
            "https://example.com?foo=bar",
        )

    def test_add_url_params_preserves_existing(self):
        url = "https://example.com?one=1"
        self.assertEqual(
            _add_params_to_url_if_new(url, {"foo": "bar"}),
            "https://example.com?foo=bar&one=1",
        )

    def test_add_url_params_prefers_existing(self):
        url = "https://example.com?one=1"
        self.assertEqual(
            _add_params_to_url_if_new(url, {"one": "2"}), "https://example.com?one=1"
        )

    def test_get_visible_links_raises(self):
        with self.assertRaises(NoReverseMatch):
            get_visible_links(
                context={},
                user=user_with_perms(["testapp.view_dragonfly"]),
                links=DragonflyViewSet().links,
                link_layout=["detail"],
                obj=None,
            )

    def test_get_visible_links_checks_permission(self):
        links = get_visible_links(
            context={},
            user=user_with_perms(["testapp.view_dragonfly", "testapp.add_dragonfly"]),
            links=DragonflyViewSet().links,
            link_layout=["detail", "update", "create"],
            obj=Dragonfly(pk=123),
        )
        self.assertEqual(
            {component.name for component, url in links}, {"create", "detail"}
        )

    def test_apps_for_navigation(self):
        request = RequestFactory().get("/")
        request.user = user_with_perms(
            ["testapp.view_dragonfly", "testapp.view_sighting"]
        )
        context = RequestContext(request, {"request": request})
        apps = get_apps_for_navigation(context=context)
        self.assertEqual(len(apps), 1)
        self.assertEqual(apps[0]["app_label"], "testapp")
        self.assertEqual(
            apps[0]["entries"],
            [("dragonflys", "/dragonfly/"), ("sightings", "/sighting/")],
        )

    def test_apps_for_navigation_require_permission(self):
        request = RequestFactory().get("/")
        request.user = user_with_perms([])
        context = RequestContext(request, {"request": request})
        apps = get_apps_for_navigation(context=context)
        self.assertEqual(apps, [])



from typing import List, Optional, Type

from django.contrib.admin.utils import NestedObjects
from django.core.exceptions import ValidationError
from django.core.paginator import Page, Paginator
from django.db import router
from django.db.models import Model
from django.db.models.options import Options
from django.forms import ModelForm, inlineformset_factory
from django.utils.text import get_text_list
from django.utils.translation import gettext as _
from django_filters.filterset import filterset_factory

from beam.actions import Action
from beam.types import LayoutType

DELETION_FIELD_NAME = "DELETE"


class BaseRelatedInline(object):
    model: Model
    title: str = ""
    foreign_key_field: str
    layout: Optional[LayoutType] = None
    fields: List[str] = []
    extra = None
    can_delete = True
    can_order = False
    order_field: str = ""
    queryset = None
    form_class = ModelForm
    detail_template_name = ""
    form_template_name = ""

    def __init__(self, parent_instance=None, parent_model=None, request=None) -> None:
        super().__init__()
        self.parent_instance = parent_instance
        self.parent_model = parent_model
        self.request = request
        self._formset = None
        self.prefix = self.get_prefix()
        assert self.foreign_key_field, "you must set a foreign key field"
        assert (
            not self.can_order or self.order_field
        ), "you must set order_field when using can_order=True"

    @property
    def formset(self):
        if self._formset is None:
            self._formset = self.construct_formset()
        return self._formset

    def get_prefix(self):
        # FIXME would rather have a unique prefix!
        # same implementation as default InlineFormset
        return (
            self.model_options.get_field(self.foreign_key_field)
            .remote_field.get_accessor_name(model=self.model)
            .replace("+", "")
        )

    def _construct_form_class(self):
        """
        Build a form class based on self.form_class that handles deletion of nested objects.

        Implementation from django.contrib.admin.options.get_formset.
        """

        class DeleteProtectedModelForm(self.form_class):
            def hand_clean_DELETE(self):
                """
                We don't validate the 'DELETE' field itself because on
                templates it's not rendered using the field information.
                """
                if self.cleaned_data.get(DELETION_FIELD_NAME, False):
                    using = router.db_for_write(self._meta.model)
                    collector = NestedObjects(using=using)
                    if self.instance._state.adding:
                        return
                    collector.collect([self.instance])
                    if collector.protected:
                        objs = []
                        for p in collector.protected:
                            objs.append(
                                # Translators: Model verbose name and instance representation,
                                # suitable to be an item in a list.
                                _("{class_name} {instance}").format(
                                    class_name=p._meta.verbose_name, instance=p
                                )
                            )
                        msg = _(
                            "Deleting {class_name} {instance} would require "
                            "deleting the following protected related objects: "
                            "{related_objects}"
                        ).format(
                            class_name=self._meta.model._meta.verbose_name,
                            instance=self.instance,
                            related_objects=get_text_list(objs, _("and")),
                        )
                        raise ValidationError(msg, code="deleting_protected")

            def clean(self):
                cleaned_data = super().clean() or self.cleaned_data
                try:
                    self.hand_clean_DELETE()
                except ValidationError as e:
                    cleaned_data.pop(DELETION_FIELD_NAME)
                    raise e
                return cleaned_data

        return DeleteProtectedModelForm

    def get_formset_class(self):
        if self.extra is not None:
            extra = self.extra
        elif self.parent_instance:
            extra = 0
        else:
            extra = 1

        return inlineformset_factory(
            parent_model=self.parent_model,
            form=self._construct_form_class(),
            model=self.model,
            fk_name=self.foreign_key_field,
            extra=extra,
            can_delete=self.can_delete,
            fields=self.fields[:],
        )

    def get_formset_kwargs(self):
        return {
            "instance": self.parent_instance,
            "data": self.request.POST if self.request and self.request.POST else None,
            "files": self.request.FILES
            if self.request and self.request.FILES
            else None,
            "prefix": self.prefix,
        }

    def construct_formset(self):
        formset_class = self.get_formset_class()
        return formset_class(**self.get_formset_kwargs())

    def get_title(self):
        return self.title or self.model_options.verbose_name_plural

    def get_queryset(self):
        if self.queryset:
            # ensure re-evaluation of queryset
            qs = self.queryset.all()
        else:
            qs = self._get_queryset_from_parent_instance()

        if not qs.ordered:
            # ensure a stable order
            qs = qs.order_by("pk")

        return qs

    def _get_queryset_from_parent_instance(self):
        if self.parent_instance is None or self.parent_instance.pk is None:
            return self.model.objects.none()
        related_name = self.model_options.get_field(
            self.foreign_key_field
        ).remote_field.get_accessor_name()
        return getattr(self.parent_instance, related_name).all()

    @property
    def model_options(self) -> Options:
        return self.model._meta


class NotPaginated(Page):
    """
    A single page that holds all objects.
    """

    def __init__(self, object_list, number, paginator):
        super().__init__(object_list, number, paginator)

        class NoPaginator:
            @property
            def count(self):
                return len(object_list)

        self.paginator = NoPaginator()

    def has_next(self):
        return False

    def has_previous(self):
        return False

    def start_index(self):
        return 1

    def end_index(self):
        return 1


class PaginationMixin(BaseRelatedInline):
    paginate_by: Optional[int] = None
    paginator_class = Paginator

    def __init__(self, parent_instance=None, parent_model=None, request=None) -> None:
        super().__init__(parent_instance, parent_model, request)
        assert not (
            self.paginate_by and self.can_order
        ), "ordering can't be enabled at the same time as pagination"

    @property
    def page_param(self):
        return "{}-page".format(self.prefix)

    @property
    def page(self) -> Page:
        queryset = self.get_queryset()
        if self.paginate_by:
            paginator = self.paginator_class(queryset, self.paginate_by)
            page_number = self.request.GET.get(self.page_param, 1)
            return paginator.page(page_number)
        else:
            return NotPaginated(object_list=queryset, number=1, paginator=None)

    def get_formset_kwargs(self):
        kwargs = super().get_formset_kwargs()
        if self.paginate_by:
            queryset = self.get_queryset()
            # because pagination will slice the queryset and the formset will filter
            # it again which isn't possible so we create an unsliced one with just
            # the objects on the current page
            object_ids = (obj.pk for obj in self.page.object_list)
            kwargs["queryset"] = queryset.filter(pk__in=object_ids)
        return kwargs


class FilterSetMixin(BaseRelatedInline):
    filterset_class = None
    filterset_fields: Optional[List[str]] = None

    def __init__(self, parent_instance=None, parent_model=None, request=None) -> None:
        super().__init__(parent_instance, parent_model, request)
        self.filterset = self.get_filterset()

    def get_filterset_fields(self):
        return self.filterset_fields

    def get_filterset_class(self):
        if self.filterset_class:
            return self.filterset_class
        elif self.get_filterset_fields():
            return filterset_factory(
                model=self.model, fields=self.get_filterset_fields()
            )
        return None

    def get_filterset_kwargs(self):
        """
        Returns the keyword arguments for instantiating the filterset.
        """

        kwargs = {
            "data": self.request.GET or None,
            "request": self.request,
            # self.get_queryset() uses filterset to filter so must use super()
            "queryset": super().get_queryset(),
            "prefix": "{}-filter".format(self.prefix),
        }
        return kwargs

    def get_filterset(self):
        filterset_class = self.get_filterset_class()
        if not filterset_class:
            return None
        return filterset_class(**self.get_filterset_kwargs())

    def get_queryset(self):
        if self.filterset and self.filterset.is_bound and self.filterset.is_valid():
            return self.filterset.qs
        else:
            return super().get_queryset()


class ActionsMixin(BaseRelatedInline):
    action_classes: List[Type[Action]] = []

    def __init__(self, parent_instance=None, parent_model=None, request=None) -> None:
        super().__init__(parent_instance, parent_model, request)
        self.actions = self.get_actions()

    def get_action(self):
        for action in self.actions:
            if action.is_bound:
                return action
        return None

    def get_actions(self):
        if not self.action_classes:
            return []

        selected_action = self.request.POST.get("_action_choice")
        actions = []
        action_class: Type[Action]
        for index, action_class in enumerate(self.action_classes):
            action_id = "{}-{}-{}".format(self.prefix, index, action_class.name)
            action = action_class(
                data=self.request.POST if action_id == selected_action else None,
                model=self.model,
                id=action_id,
                request=self.request,
            )
            if action.has_perm(self.request.user):
                actions.append(action)
        return actions


class RelatedInline(ActionsMixin, FilterSetMixin, PaginationMixin, BaseRelatedInline):
    pass


class TabularRelatedInline(RelatedInline):
    detail_template_name = "beam/partials/detail_inline_tabular.html"
    form_template_name = "beam/partials/form_inline_tabular.html"



from beam.actions import DeleteAction
from django.test import TestCase
from django_webtest import WebTest
from test_views import user_with_perms
from testapp.models import Dragonfly, ProtectedSighting, Sighting
from testapp.views import DragonFlyUpdateAction, DragonflyViewSet


class ActionTest(TestCase):
    def test_delete_action(self):
        action = DeleteAction(
            data=None, model=Dragonfly, id="delete-dragonfly", request=None
        )
        Dragonfly.objects.create(name="alpha", age=10)
        Dragonfly.objects.create(name="beta", age=10)
        Dragonfly.objects.create(name="gamma", age=100)
        action.apply(Dragonfly.objects.filter(age__lte=10))
        self.assertEqual(action.get_success_message(), "Deleted 2 dragonflys")
        self.assertEqual(Dragonfly.objects.count(), 1)

    def test_mass_update_action(self):
        Dragonfly.objects.create(name="alpha", age=10)
        Dragonfly.objects.create(name="beta", age=10)
        Dragonfly.objects.create(name="gamma", age=100)

        action = DragonFlyUpdateAction(
            data={"update-age": "100"}, model=Dragonfly, id="update", request=None
        )
        action.apply(Dragonfly.objects.all())

        self.assertEqual(action.get_success_message(), "Updated 2 dragonflys")
        self.assertEqual(Dragonfly.objects.filter(age=100).count(), 3)


class ActionViewTest(WebTest):
    def setUp(self):
        self.alpha = Dragonfly.objects.create(name="alpha", age=12)
        self.omega = Dragonfly.objects.create(name="omega", age=99)

    def test_no_action_selected_does_nothing(self):
        list_page = self.app.get(
            DragonflyViewSet().links["list"].reverse(),
            user=user_with_perms(
                ["testapp.view_dragonfly", "testapp.delete_dragonfly",]
            ),
        )
        form = list_page.forms["list-action-form"]
        form["_action_select[]"] = [self.alpha.pk]
        response = form.submit()
        self.assertContains(response, "alpha")
        self.assertContains(response, "omega")
        self.assertEqual(Dragonfly.objects.count(), 2)

    def test_no_list_items_selected_does_nothing(self):
        list_page = self.app.get(
            DragonflyViewSet().links["list"].reverse(),
            user=user_with_perms(
                ["testapp.view_dragonfly", "testapp.delete_dragonfly",]
            ),
        )
        form = list_page.forms["list-action-form"]
        form["_action_choice"] = "0-delete"
        response = form.submit().follow()
        self.assertContains(response, "alpha")
        self.assertContains(response, "omega")
        self.assertEqual(Dragonfly.objects.count(), 2)

    def test_action_with_selection(self):
        list_page = self.app.get(
            DragonflyViewSet().links["list"].reverse(),
            user=user_with_perms(
                ["testapp.view_dragonfly", "testapp.delete_dragonfly",]
            ),
        )
        form = list_page.forms["list-action-form"]
        form["_action_choice"] = "0-delete"
        form["_action_select[]"] = [self.alpha.pk]
        response = form.submit().follow()

        self.assertNotContains(response, "alpha")
        self.assertContains(response, "omega")

        self.assertEqual(Dragonfly.objects.count(), 1)

    def test_selection_across_page_boundary(self):
        list_page = self.app.get(
            DragonflyViewSet().links["list"].reverse(),
            user=user_with_perms(
                ["testapp.view_dragonfly", "testapp.delete_dragonfly",]
            ),
        )
        form = list_page.forms["list-action-form"]
        form["_action_choice"] = "0-delete"
        form["_action_select_across"] = "all"
        response = form.submit().follow()

        self.assertNotContains(response, "alpha")
        self.assertNotContains(response, "omega")

        self.assertEqual(Dragonfly.objects.count(), 0)

    def test_no_permission_hides_action_from_menu(self):
        list_page = self.app.get(
            DragonflyViewSet().links["list"].reverse(),
            user=user_with_perms(
                ["testapp.view_dragonfly", "testapp.change_dragonfly",]
            ),
        )
        form = list_page.forms["list-action-form"]
        with self.assertRaises(ValueError):
            form["_action_choice"] = "0-delete"

    def test_no_permission_prevents_action(self):
        list_page = self.app.get(
            DragonflyViewSet().links["list"].reverse(),
            user=user_with_perms(
                ["testapp.view_dragonfly", "testapp.change_dragonfly",]
            ),
        )
        form = list_page.forms["list-action-form"]
        form["_action_choice"].force_value("0-delete")
        form["_action_select_across"] = "all"
        response = form.submit()

        self.assertContains(response, "alpha")
        self.assertContains(response, "omega")

        self.assertEqual(Dragonfly.objects.count(), 2)

    def test_invalid_form_shows_error_and_preserves_action_choice(self):
        list_page = self.app.get(
            DragonflyViewSet().links["list"].reverse(),
            user=user_with_perms(
                ["testapp.view_dragonfly", "testapp.change_dragonfly",]
            ),
        )
        form = list_page.forms["list-action-form"]
        form["_action_choice"] = "1-update_selected"
        form["_action_select[]"] = [self.alpha.pk]
        form["1-update_selected-age"] = "not_a_number"

        response = form.submit()

        self.assertContains(response, "Enter a whole number")
        self.assertEqual(
            response.forms["list-action-form"]["_action_choice"].value,
            "1-update_selected",
        )

    def test_filter_interaction(self):
        list_page = self.app.get(
            DragonflyViewSet().links["list"].reverse() + "?filter-name=alpha",
            user=user_with_perms(
                ["testapp.view_dragonfly", "testapp.delete_dragonfly",]
            ),
        )
        form = list_page.forms["list-action-form"]
        form["_action_choice"] = "0-delete"
        form["_action_select_across"] = "all"
        response = form.submit().follow()

        self.assertContains(response, "Deleted 1 dragonflys")

        self.assertFalse(Dragonfly.objects.filter(name="alpha").exists())
        self.assertTrue(Dragonfly.objects.filter(name="omega").exists())

    def test_search_interaction(self):
        list_page = self.app.get(
            DragonflyViewSet().links["list"].reverse() + "?q=omega",
            user=user_with_perms(
                ["testapp.view_dragonfly", "testapp.delete_dragonfly",]
            ),
        )
        form = list_page.forms["list-action-form"]
        form["_action_choice"] = "0-delete"
        form["_action_select_across"] = "all"
        response = form.submit().follow()

        self.assertContains(response, "Deleted 1 dragonflys")

        self.assertTrue(Dragonfly.objects.filter(name="alpha").exists())
        self.assertFalse(Dragonfly.objects.filter(name="omega").exists())


class InlineActionViewTest(WebTest):
    def setUp(self):
        self.dragonfly = Dragonfly.objects.create(name="alpha", age=12)
        Sighting.objects.create(dragonfly=self.dragonfly, name="regular-sighting-0")
        Sighting.objects.create(dragonfly=self.dragonfly, name="regular-sighting-1")
        ProtectedSighting.objects.create(
            dragonfly=self.dragonfly, name="protected-sighting-0"
        )

    def test_action_shown_only_where_permission(self):
        detail_page = self.app.get(
            DragonflyViewSet().links["detail"].reverse(self.dragonfly),
            user=user_with_perms(
                [
                    "testapp.view_dragonfly",
                    "testapp.delete_sighting",
                    "testapp.delete_protectedsighting",
                ],
                username="user-0",
            ),
        )
        self.assertContains(detail_page, "sighting_set-0-delete")
        self.assertContains(detail_page, "protectedsighting_set-0-delete")

        detail_page_only_sighting = self.app.get(
            DragonflyViewSet().links["detail"].reverse(self.dragonfly),
            user=user_with_perms(
                ["testapp.view_dragonfly", "testapp.delete_sighting",],
                username="user-1",
            ),
        )
        self.assertContains(detail_page_only_sighting, "sighting_set-0-delete")
        self.assertNotContains(
            detail_page_only_sighting, "protectedsighting_set-0-delete"
        )

        detail_page_only_protected_sighting = self.app.get(
            DragonflyViewSet().links["detail"].reverse(self.dragonfly),
            user=user_with_perms(
                ["testapp.view_dragonfly", "testapp.delete_protectedsighting",],
                username="user-2",
            ),
        )
        self.assertContains(
            detail_page_only_protected_sighting, "sighting_set-0-delete"
        )
        self.assertContains(
            detail_page_only_protected_sighting, "protectedsighting_set-0-delete"
        )

    def test_action_with_selection(self):
        detail_page = self.app.get(
            DragonflyViewSet().links["detail"].reverse(self.dragonfly),
            user=user_with_perms(
                ["testapp.view_dragonfly", "testapp.delete_sighting",]
            ),
        )
        form = detail_page.forms["sighting_set-action-form"]
        form["_action_choice"] = "sighting_set-0-delete"
        form["_action_select[]"] = [Sighting.objects.get(name="regular-sighting-0").pk]
        response = form.submit().follow()
        self.assertContains(response, "Deleted 1 sighting")
        self.assertNotContains(response, "regular-sighting-0")
        self.assertContains(response, "regular-sighting-1")

        self.assertEqual(Sighting.objects.count(), 1)

    def test_selection_across_page_boundary(self):
        Sighting.objects.create(dragonfly=self.dragonfly, name="regular-sighting-2")
        Sighting.objects.create(dragonfly=self.dragonfly, name="regular-sighting-3")
        Sighting.objects.create(dragonfly=self.dragonfly, name="regular-sighting-4")
        Sighting.objects.create(dragonfly=self.dragonfly, name="regular-sighting-5")
        detail_page = self.app.get(
            DragonflyViewSet().links["detail"].reverse(self.dragonfly),
            user=user_with_perms(
                ["testapp.view_dragonfly", "testapp.delete_sighting",]
            ),
        )
        form = detail_page.forms["sighting_set-action-form"]
        form["_action_choice"] = "sighting_set-0-delete"
        form["_action_select_across"] = "all"
        response = form.submit().follow()

        self.assertNotContains(response, "regular-sighting-0")
        self.assertNotContains(response, "regular-sighting-1")

        self.assertEqual(Sighting.objects.count(), 0)

    def test_mass_update_form(self):
        detail_page = self.app.get(
            DragonflyViewSet().links["detail"].reverse(self.dragonfly),
            user=user_with_perms(
                [
                    "testapp.view_dragonfly",
                    "testapp.delete_sighting",
                    "testapp.change_sighting",
                ]
            ),
        )
        form = detail_page.forms["sighting_set-action-form"]
        form["_action_choice"] = "sighting_set-1-update_selected"
        form["_action_select_across"] = "all"
        form["sighting_set-1-update_selected-name"] = "a new name"
        response = form.submit().follow()

        self.assertNotContains(response, "regular-sighting-0")
        self.assertEqual(Sighting.objects.filter(name="a new name").count(), 2)
        self.assertEqual(Sighting.objects.exclude(name="a new name").count(), 0)

    def test_no_interaction_with_multiple_inlines(self):
        self.dragonfly.sighting_set.create(name="interaction-test-sighting", pk=999)
        self.dragonfly.protectedsighting_set.create(
            name="interaction-test-protected-sighting", pk=999
        )
        detail_page = self.app.get(
            DragonflyViewSet().links["detail"].reverse(self.dragonfly),
            user=user_with_perms(
                [
                    "testapp.view_dragonfly",
                    "testapp.delete_sighting",
                    "testapp.delete_protectedsighting",
                ]
            ),
        )
        self.assertTrue(Sighting.objects.filter(pk=999).exists())
        self.assertTrue(ProtectedSighting.objects.filter(pk=999).exists())

        sighting_form = detail_page.forms["sighting_set-action-form"]
        sighting_form["_action_choice"] = "sighting_set-0-delete"
        sighting_form["_action_select[]"] = [999]
        sighting_response = sighting_form.submit().follow()
        self.assertContains(sighting_response, "Deleted 1 sighting")

        self.assertFalse(Sighting.objects.filter(pk=999).exists())
        self.assertTrue(ProtectedSighting.objects.filter(pk=999).exists())

        protected_sighting_form = detail_page.forms["protectedsighting_set-action-form"]
        protected_sighting_form["_action_choice"] = "protectedsighting_set-0-delete"
        protected_sighting_form["_action_select[]"] = ["999"]
        protected_sighting_response = protected_sighting_form.submit().follow()
        self.assertContains(protected_sighting_response, "Deleted 1 protected sighting")

        self.assertFalse(Sighting.objects.filter(pk=999).exists())
        self.assertFalse(ProtectedSighting.objects.filter(pk=999).exists())

    def test_filter_interaction(self):
        user = user_with_perms(["testapp.view_dragonfly", "testapp.delete_sighting"],)

        detail_page = self.app.get(
            DragonflyViewSet().links["detail"].reverse(self.dragonfly), user=user,
        )

        self.assertContains(detail_page, "regular-sighting-0")
        self.assertContains(detail_page, "regular-sighting-1")

        filter_form = detail_page.forms["sighting_set-filter-form"]
        filter_form["sighting_set-filter-name"] = "regular-sighting-0"

        filtered_page = filter_form.submit()

        self.assertContains(filtered_page, "regular-sighting-0")
        self.assertNotContains(filtered_page, "regular-sighting-1")

        action_form = filtered_page.forms["sighting_set-action-form"]
        action_form["_action_choice"] = "sighting_set-0-delete"
        action_form["_action_select_across"] = "all"
        action_response = action_form.submit().follow()

        self.assertContains(action_response, "Deleted 1 sighting")

        self.assertFalse(Sighting.objects.filter(name="regular-sighting-0").exists())
        self.assertTrue(Sighting.objects.filter(name="regular-sighting-1").exists())

    def test_http_response_from_action(self):
        user = user_with_perms(["testapp.view_dragonfly", "testapp.view_sighting"],)

        detail_page = self.app.get(
            DragonflyViewSet().links["detail"].reverse(self.dragonfly), user=user,
        )

        form = detail_page.forms["sighting_set-action-form"]
        form["_action_choice"] = "sighting_set-2-csv_export"
        form["_action_select_across"] = "all"
        response = form.submit()
        self.assertEqual(
            response.content.decode("utf-8"),
            "{},regular-sighting-0\r\n"
            "{},regular-sighting-1\r\n".format(
                Sighting.objects.get(name="regular-sighting-0").pk,
                Sighting.objects.get(name="regular-sighting-1").pk,
            ),
        )



from typing import Dict
from unittest.mock import Mock

from beam.components import BaseComponent, Component
from beam.viewsets import BaseViewSet, undefined
from django.test import TestCase


class ViewsetTest(TestCase):
    def test_get_components(self):
        class ViewSet(BaseViewSet):
            model = Mock()
            view_class = Mock()

            def get_component_classes(self):
                return super().get_component_classes() + [("test", Component)]

        self.assertEqual(len(ViewSet().components), 1)

    def test_links(self):
        class ViewSet(BaseViewSet):
            model = Mock()
            view_class = Mock()

            def get_component_classes(self):
                return super().get_component_classes() + [("test", Component)]

            @property
            def links(self) -> Dict[str, BaseComponent]:
                super_links = super().links
                super_links["extra"] = Mock()
                return super_links

        links = ViewSet().links
        assert len(links) == 2
        self.assertEqual(len(links), 2)
        self.assertIn("extra", links)
        self.assertIn("test", links)

    def test_get_urls(self):
        class ViewSet(BaseViewSet):
            model = Mock()
            view_class = Mock()

            def get_component_classes(self):
                return super().get_component_classes() + [("test", Component)]

            url = "foo/<pk>"

        urls = ViewSet().get_urls()

        self.assertEqual(len(urls), 1)
        self.assertEqual(urls[0].pattern._route, "foo/<pk>")

    def test_get_view(self):
        class ViewSet(BaseViewSet):
            model = Mock()
            view_class = Mock()

            def get_component_classes(self):
                return super().get_component_classes() + [("test", Component)]

            url = "foo/<pk>"

        viewset = ViewSet()
        component = viewset.components["test"]

        self.assertEqual(
            viewset._get_view(component).__wrapped__, ViewSet.view_class.as_view()
        )

    def test_resolve_component_kwargs(self):
        class ViewSet(BaseViewSet):
            model = Mock()
            view_class = Mock()
            arg = "value"

        self.assertEqual(
            ViewSet()._resolve_component_kwargs("test", ["arg"])["arg"], "value"
        )

    def test_resolve_component_kwargs_does_not_return_unspecified_value(self):
        class ViewSet(BaseViewSet):
            model = Mock()
            view_class = Mock()

        component = Mock()
        component.get_arguments.return_value = ["arg"]

        self.assertNotIn("arg", ViewSet()._resolve_component_kwargs("test", ["arg"]))

    def test_resolve_component_kwargs_does_not_return_sentinel_value(self):
        class ViewSet(BaseViewSet):
            model = Mock()
            view_class = Mock()
            arg = undefined

        self.assertNotIn("arg", ViewSet()._resolve_component_kwargs("test", ["arg"]))

    def test_resolve_component_kwargs_does_return_falsy_values(self):
        class ViewSet(BaseViewSet):
            model = Mock()
            view_class = Mock()
            arg = None

        self.assertIsNone(ViewSet()._resolve_component_kwargs("test", ["arg"])["arg"])

    def test_resolve_component_kwargs_prefers_specific(self):
        class ViewSet(BaseViewSet):
            model = Mock()
            view_class = Mock()
            arg = "value"
            test_arg = "specific_value"

        self.assertEqual(
            ViewSet()._resolve_component_kwargs("test", ["arg"])["arg"],
            "specific_value",
        )

    def test_resolve_component_kwargs_does_not_return_specific_undefined(self):
        class ViewSet(BaseViewSet):
            model = Mock()
            view_class = Mock()
            arg = "value"
            test_arg = undefined

        self.assertEqual(
            ViewSet()._resolve_component_kwargs("test", ["arg"])["arg"], "value"
        )

    def test_resolve_component_kwargs_prefers_specific_falsy(self):
        class ViewSet(BaseViewSet):
            model = Mock()
            view_class = Mock()
            arg = "value"
            test_arg = None

        self.assertIsNone(ViewSet()._resolve_component_kwargs("test", ["arg"])["arg"])



from beam import ViewSet
from beam.contrib.autocomplete_light import AutocompleteMixin
from beam.registry import RegistryType
from django.core.exceptions import PermissionDenied
from django.test import RequestFactory, TestCase
from test_views import user_with_perms
from testapp.models import Dragonfly

registry: RegistryType = {}


class AutocompleteDragonflyViewSet(AutocompleteMixin, ViewSet):
    registry = registry

    model = Dragonfly
    fields = ["name", "age"]
    autocomplete_search_fields = ["name"]


class AutocompleteTest(TestCase):
    def test_autocomplete(self):
        request = RequestFactory().get("/", {})
        request.user = user_with_perms(["testapp.view_dragonfly"])

        Dragonfly.objects.create(name="alpha", age=12)
        Dragonfly.objects.create(name="omega", age=99)

        view = AutocompleteDragonflyViewSet()._get_view(
            AutocompleteDragonflyViewSet().components["autocomplete"]
        )

        response = view(request)

        self.assertContains(response, "alpha")
        self.assertContains(response, "omega")

    def test_autocomplete_search(self):
        request = RequestFactory().get("/", {"q": "Al"})
        request.user = user_with_perms(["testapp.view_dragonfly"])

        Dragonfly.objects.create(name="alpha", age=12)
        Dragonfly.objects.create(name="omega", age=99)

        view = AutocompleteDragonflyViewSet()._get_view(
            AutocompleteDragonflyViewSet().components["autocomplete"]
        )

        response = view(request)

        self.assertContains(response, "alpha")
        self.assertNotContains(response, "omega")

    def test_autocomplete_requires_permission(self):
        Dragonfly.objects.create(name="alpha", age=47)

        request = RequestFactory().get("/", {})
        request.user = user_with_perms([])

        view = AutocompleteDragonflyViewSet()._get_view(
            AutocompleteDragonflyViewSet().components["autocomplete"]
        )

        with self.assertRaises(PermissionDenied):
            view(request)



from unittest import mock

from django.test import TestCase
from django.test.utils import override_settings
from django.urls import NoReverseMatch, include, path, reverse
from testapp.models import Dragonfly
from testapp.views import DragonflyViewSet, SightingViewSet

from beam.templatetags.beam_tags import get_link_url, get_url_for_related

urlpatterns = [
    path(
        "dragonfly/",
        include(
            (DragonflyViewSet().get_urls(), "my_app_name"), namespace="my_namespace"
        ),
    ),
    path(
        "sighting/",
        include(
            (SightingViewSet().get_urls(), "my_app_name"), namespace="my_namespace"
        ),
    ),
]


class UrlTest(TestCase):
    def test_get_urls_produces_urls(self):
        self.assertEqual(len(DragonflyViewSet().get_urls()), 6)

    def test_get_links_contains_all_view_types(self):
        self.assertSetEqual(
            set(DragonflyViewSet().links.keys()),
            {"list", "detail", "update", "create", "delete", "extra"},
        )

    def test_url_patterns_are_correct(self):
        expected = {
            "testapp_dragonfly_list": "",
            "testapp_dragonfly_create": "create/",
            "testapp_dragonfly_detail": "<str:pk>/",
            "testapp_dragonfly_delete": "<str:pk>/delete/",
            "testapp_dragonfly_update": "<str:pk>/update/",
            "testapp_dragonfly_extra": "extra/<str:pk>/<str:special>/",
        }

        for url in DragonflyViewSet().get_urls():
            self.assertEqual(str(url.pattern), expected[url.name])

    def test_links_that_do_not_require_an_instance(self):
        links = DragonflyViewSet().links
        self.assertEqual(get_link_url(None, links["list"]), "/dragonfly/")
        self.assertEqual(get_link_url(None, links["create"]), "/dragonfly/create/")

    def test_links_that_do_not_require_an_instance_work_if_one_is_supplied(self):
        links = DragonflyViewSet().links
        instance = Dragonfly(pk=123)
        self.assertEqual(get_link_url(None, links["list"], instance), "/dragonfly/")
        self.assertEqual(
            get_link_url(None, links["create"], instance), "/dragonfly/create/"
        )

    def test_links_that_require_an_instance(self):
        links = DragonflyViewSet().links
        instance = Dragonfly(pk=123)
        self.assertEqual(
            get_link_url(None, links["detail"], instance), "/dragonfly/123/"
        )
        self.assertEqual(
            get_link_url(None, links["update"], instance), "/dragonfly/123/update/"
        )
        self.assertEqual(
            get_link_url(None, links["delete"], instance), "/dragonfly/123/delete/"
        )

    def test_links_that_require_an_instance_raise_if_missing(self):
        links = DragonflyViewSet().links
        with self.assertRaises(NoReverseMatch):
            get_link_url(None, links["detail"], None)
        with self.assertRaises(NoReverseMatch):
            get_link_url(None, links["update"], None)
        with self.assertRaises(NoReverseMatch):
            get_link_url(None, links["delete"], None)

    def test_component_url_with_extra_context(self):
        instance = Dragonfly(pk=123)
        link = DragonflyViewSet().links["extra"]
        self.assertEqual(
            get_link_url(None, link, instance, special="param"),
            "/dragonfly/extra/123/param/",
        )

    def test_component_url_kwarg_from_request(self):
        class request:
            class resolver_match:
                kwargs = {"special": "from_request"}

        instance = Dragonfly(pk=123)
        url = DragonflyViewSet().links["extra"].reverse(instance, request)
        self.assertEqual(
            url,
            "/dragonfly/extra/123/from_request/",
        )

    def test_get_url_for_related_raises_if_missing_params(self):
        instance = Dragonfly(pk=123)
        with self.assertRaises(NoReverseMatch):
            get_url_for_related({}, instance, "extra")

    @override_settings(ROOT_URLCONF=__name__)
    @mock.patch.object(DragonflyViewSet, "url_namespace", "my_namespace")
    def test_namespace_urls(self):
        links = DragonflyViewSet().links
        instance = Dragonfly(pk=123)
        self.assertEqual(
            get_link_url(None, links["detail"], instance), "/dragonfly/123/"
        )
        self.assertEqual(
            get_link_url(None, links["update"], instance), "/dragonfly/123/update/"
        )
        self.assertEqual(
            get_link_url(None, links["delete"], instance), "/dragonfly/123/delete/"
        )
        self.assertEqual(get_link_url(None, links["list"]), "/dragonfly/")
        self.assertEqual(get_link_url(None, links["create"]), "/dragonfly/create/")

        self.assertEqual(reverse("my_namespace:testapp_dragonfly_list"), "/dragonfly/")



from unittest import TestCase
from unittest.mock import Mock

from testapp.views import DragonflyViewSet

from beam.components import BaseComponent, Component


class ComponentTest(TestCase):
    def test_component_with_queryset_guesses_model(self):
        queryset = Mock()
        assert (
            Component(name="test", queryset=queryset, view_class=Mock()).model
            == queryset.model
        )

    def test_component_with_queryset_copies_queryset(self):
        queryset = Mock()
        assert (
            Component(name="test", queryset=queryset, view_class=Mock()).queryset
            is not queryset
        )

    def test_component_with_model_guesses_queryset(self):
        model = Mock()
        assert (
            Component(name="test", model=model, view_class=Mock()).queryset
            == model._default_manager.all()
        )

    def test_component_without_verbose_name_guesses_name(self):
        model = Mock()
        assert (
            Component(name="test", model=model, view_class=Mock()).queryset
            == model._default_manager.all()
        )

    def test_base_component_requires_a_name(self):
        with self.assertRaises(ValueError):
            BaseComponent()
        BaseComponent(name="foo")

    def test_component_requires_one_of_model_or_queryset(self):
        with self.assertRaises(ValueError):
            Component(name="foo", view_class=Mock())
        Component(name="foo", model=Mock(), view_class=Mock())
        Component(name="foo", queryset=Mock(), view_class=Mock())

    def test_component_requires_a_view_class(self):
        with self.assertRaises(ValueError):
            Component(name="foo", model=Mock())
        Component(name="foo", model=Mock(), view_class=Mock())

    def test_arguments_are_collected(self):
        class SubComponent(Component):
            def __init__(self, new_arg_1=None, **kwargs):
                super().__init__(**kwargs)

        class SubSubComponent(SubComponent):
            def __init__(self, new_arg_2=None, **kwargs):
                super().__init__(**kwargs)

        assert "new_arg_1" in SubComponent.get_arguments()
        assert "new_arg_2" not in SubComponent.get_arguments()

        assert "new_arg_1" in SubSubComponent.get_arguments()
        assert "new_arg_2" in SubSubComponent.get_arguments()

    def test_component_has_a_sensible_string_representation(self):
        self.assertEqual(
            str(DragonflyViewSet().links["detail"]),
            "<Component testapp.views.DragonflyViewSet 'detail'>",
        )

    def test_component_has_a_sensible_string_representation_without_viewset(self):
        self.assertEqual(
            str(BaseComponent(viewset=None, name="detail")),
            "<BaseComponent 'detail'>",
        )
